try:
    from .csp import *
    from .fastapi import query_json
    from .id_generator import get_counter
    from .struct import (
        GatewayLookupMixin,
        GatewayPydanticMixin,
        GatewayStruct,
        GatewayStructMixins,
        IdType,
        PerspectiveUtilityMixin,
        is_gateway_struct_like,
    )
    from .web.controls import Controls
except ImportError:
    pass

from .enums import *
from .exceptions import *
from .picklable_queue import PickleableQueue
from .threads import get_thread
from .web.filter import Filter, FilterCondition, FilterWhere, FilterWhereLambdaMap
from .web.query import Query



"""
A Demo CSP Gateway application
"""

from datetime import timedelta

import csp
from csp import ts

from csp_gateway import GatewayChannels, GatewayModule, GatewayStruct

# Let's walk through a worked example.

# First, we want to define the edges that will be available to our modules.
# In `csp-gateway` dialect, these deferred edges are called `channels`.
# Let's define a few `csp.Struct` to go with them.

# NOTE: See __main__.py for the entry point to run this example

__all__ = (
    "ExampleData",
    "ExampleGatewayChannels",
    "ExampleModule",
)


class ExampleData(GatewayStruct):
    x: int


# `MyGatewayChannels` is the collection of lazily-connected `csp` edges that will be provided to all modules in our graph.
# The modules in our graph are `pydantic`-wrapped `csp` modules, that use the `channel` APIs to connect to edges.
# Let's define a simple one.


class ExampleGatewayChannels(GatewayChannels):
    example: ts[ExampleData] = None


# A `GatewayModule` has two important features.
# First, it is a typed `pydantic` model, so you can define attributes in the usual `pydantic` way.
# Second, it has a `connect` method that will be provided the `GatewayChannels` instance when the graph is eventually wired together.
# You can use `get_channel` and `set_channel` to read-from and publish-to `csp` edges, respectively.


class ExampleModule(GatewayModule):
    interval: timedelta = timedelta(seconds=1)

    # An example module that ticks some data in a struct
    @csp.node
    def subscribe(self, trigger: ts[bool]) -> ts[ExampleData]:
        with csp.state():
            last_x = 0
        if csp.ticked(trigger):
            last_x += 1
            return ExampleData(x=last_x)

    def connect(self, channels: ExampleGatewayChannels):
        # Create some CSP data streams
        data = self.subscribe(csp.timer(interval=self.interval, value=True))

        # Channels set via `set_channel`
        channels.set_channel(ExampleGatewayChannels.example, data)



import multiprocessing
import os
import signal
import socket
import sys
import time
from datetime import timedelta
from unittest.mock import patch

import csp.impl.error_handling
import pytest
import requests
from fastapi.testclient import TestClient

from csp_gateway import Gateway, GatewaySettings, MountControls, MountRestRoutes
from csp_gateway.server.demo import ExampleGatewayChannels, ExampleModule
from csp_gateway.testing import CspDieModule, LongStartModule
from csp_gateway.tests.server.gateway.test_gateway import MyBuildFailureModule

csp.impl.error_handling.set_print_full_exception_stack(True)


def test_long_startup_die_cleanly(free_port):
    with patch("os._exit") as exit_mock, patch("os.kill") as kill_mock:
        # instantiate gateway
        gateway = Gateway(
            settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
            modules=[
                # NeverDieModule(),
                # CspDieModule(),
                LongStartModule(),
                ExampleModule(),
                MountRestRoutes(force_mount_all=True),
            ],
            channels=ExampleGatewayChannels(),
        )
        try:
            gateway.start(realtime=True, rest=True, build_timeout=1)
        except RuntimeError:
            assert exit_mock.call_count == 1
            assert kill_mock.call_count == 1
            return
        raise Exception("test failed")


def test_start_with_endtime(free_port):
    @csp.graph
    def user_graph(channels: ExampleGatewayChannels) -> csp.ts[int]:
        return csp.const(1)

    # instantiate gateway
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
        modules=[
            ExampleModule(),
            MountRestRoutes(force_mount_all=True),
        ],
        channels=ExampleGatewayChannels(),
    )
    res = gateway.start(user_graph=user_graph, realtime=True, rest=True, endtime=timedelta(seconds=5))
    assert "user_graph" in res


def test_start_with_endtime_usergraph_no_return(free_port):
    @csp.graph
    def user_graph(channels: ExampleGatewayChannels):
        csp.print("here", csp.const(1))

    # instantiate gateway
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
        modules=[
            ExampleModule(),
            MountRestRoutes(force_mount_all=True),
        ],
        channels=ExampleGatewayChannels(),
    )
    res = gateway.start(user_graph=user_graph, realtime=True, rest=True, endtime=timedelta(seconds=5))
    assert "user_graph" in res


def test_start_and_then_die_with_error(free_port):
    # instantiate gateway
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
        modules=[
            ExampleModule(),
            CspDieModule(),
            MountRestRoutes(force_mount_all=True),
        ],
        channels=ExampleGatewayChannels(),
    )
    try:
        gateway.start(realtime=True, rest=True)
    except RuntimeError:
        return
    raise Exception("test failed")


@pytest.mark.skipif(sys.platform == "darwin", reason="Flaky on MacOS GHA runners")
def test_start_and_then_graph_start_error(caplog, free_port):
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
        modules=[
            ExampleModule(),
            MyBuildFailureModule(),
            MountRestRoutes(force_mount_all=True),
        ],
        channels=ExampleGatewayChannels(),
    )
    try:
        gateway.start(realtime=True, rest=True, block=False, build_timeout=timedelta(seconds=60))
    except RuntimeError:
        assert "Graph start failure" in caplog.text
        return
    raise Exception("test failed")


def test_stop_with_shutdown(free_port):
    # instantiate gateway
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=free_port),
        modules=[
            ExampleModule(),
            MountRestRoutes(force_mount_all=True),
            MountControls(),
        ],
        channels=ExampleGatewayChannels(),
    )
    gateway.start(realtime=True, rest=True, _in_test=True)
    rest_client = TestClient(gateway.web_app.get_fastapi())
    response = rest_client.post("/api/v1/controls/shutdown")
    assert response.status_code == 200

    return_data = response.json()
    assert return_data[0]["status"] == "ok"


def run_gateway(port_str):
    from csp_gateway import Gateway, GatewaySettings, MountRestRoutes
    from csp_gateway.server.demo import ExampleGatewayChannels, ExampleModule

    # instantiate gateway
    gateway = Gateway(
        settings=GatewaySettings(AUTHENTICATE=False, PORT=int(port_str)),
        modules=[
            ExampleModule(),
            MountRestRoutes(force_mount_all=True),
            MountControls(),
        ],
        channels=ExampleGatewayChannels(),
    )
    # Start gateway without test mode
    gateway.start(realtime=True, rest=True)


@pytest.mark.parametrize("signal_val", [signal.SIGINT, signal.SIGTERM, signal.SIGUSR1])
def test_signal_with_shutdown(signal_val, free_port):
    REQUEST_RETRY_TIMEOUT = 2
    AFTER_KILL_WAIT_TIME = 10
    NUM_TRIES = 10

    port_str = str(free_port)
    # URL to check if the server is up
    url = f"http://localhost:{port_str}/api/v1/state"

    # Start the gateway in another process
    p = multiprocessing.Process(target=run_gateway, args=(port_str,))
    p.start()

    # Wait for it to startup
    for idx in range(NUM_TRIES + 1):
        if idx == NUM_TRIES:
            # Unable to fully start the server
            os.kill(p.pid, signal.SIGKILL)
            p.join()
            assert False
        if not p.is_alive():
            # Process has exited
            assert p.exitcode == 0
        try:
            time.sleep(REQUEST_RETRY_TIMEOUT)
            resp = requests.get(url, timeout=1)
            resp.raise_for_status()
            break
        except (requests.HTTPError, requests.Timeout, requests.ConnectionError):
            pass

    print("Server is up")

    # Send signal to invoke shutdown
    print(f"Sending SIGNAL: {signal_val}")
    os.kill(p.pid, signal_val)

    # Wait for gateway to react to signal
    p.join(AFTER_KILL_WAIT_TIME)

    # Check if gateway shutdown with proper exit status
    assert not p.is_alive()
    assert p.exitcode == 0


@pytest.mark.skipif(sys.platform == "darwin" and "GITHUB_ACTIONS" in os.environ, reason="Skipping test on macOS in CI, works locally")
def test_shutdown_with_big_red_button(free_port):
    REQUEST_RETRY_TIMEOUT = 2
    AFTER_SHUTDOWN_WAIT_TIME = 10
    NUM_TRIES = 30
    port_str = str(free_port)
    # URL to check if the server is up
    state_url = f"http://{socket.gethostname()}:{port_str}/api/v1/state"
    shutdown_url = f"http://{socket.gethostname()}:{port_str}/api/v1/controls/shutdown"

    # Start the gateway in another process
    p = multiprocessing.Process(target=run_gateway, args=(port_str,))
    p.start()
    # Wait for it to startup
    for idx in range(NUM_TRIES + 1):
        if idx == NUM_TRIES:
            # Unable to fully start the server
            # Unable to fully start the server
            os.kill(p.pid, signal.SIGKILL)
            p.join()
            assert False
        if not p.is_alive():
            # Process has exited
            assert p.exitcode == 0
        try:
            time.sleep(REQUEST_RETRY_TIMEOUT)
            resp = requests.get(state_url, timeout=1)
            resp.raise_for_status()
            break
        except (requests.HTTPError, requests.Timeout, requests.ConnectionError):
            print("Server not up yet")
            continue
    print("Server is up")
    # Send signal to invoke shutdown
    print("Sending Shutdown Request")
    resp = requests.post(shutdown_url, timeout=1)
    # Wait for gateway to react to signal
    p.join(AFTER_SHUTDOWN_WAIT_TIME)
    # Check if gateway shutdown with proper exit status
    assert not p.is_alive()
    assert p.exitcode == 0



import logging
from datetime import datetime, timedelta

import csp
import pytest

from csp_gateway import Gateway, LogChannels
from csp_gateway.testing.shared_helpful_classes import (
    MyDictBasketModule,
    MyExampleModule,
    MyGateway,
    MyGatewayChannels,
    MyGetModule,
    MyNoTickDictBasket,
    MySetModule,
    MySmallGatewayChannels,
    MyStruct,
)


@pytest.mark.parametrize("which_channels", [[MySmallGatewayChannels.example], {}])
def test_LogChannels(which_channels, caplog):
    logger = LogChannels(selection=which_channels)
    gateway = Gateway(
        modules=[MyExampleModule(), logger],
        channels=MySmallGatewayChannels(),
    )
    with caplog.at_level(logging.INFO):
        gateway.start(starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=1))
    expected = (
        "csp_gateway.server.modules.logging.logging",
        logging.INFO,
        "2020-01-01 00:00:01 example:1",
    )
    assert expected in caplog.record_tuples


def test_LogChannels_dict_basket(caplog):
    logger = LogChannels(log_name="gateway_modules_test_logger")
    gateway = Gateway(
        modules=[
            MyDictBasketModule(my_data=csp.const(2.0)),
            logger,
        ],
        channels=MySmallGatewayChannels(),
    )
    with caplog.at_level(logging.INFO, logger="gateway_modules_test_logger"):
        gateway.start(starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=1))
    expected = (
        "gateway_modules_test_logger",
        20,
        "2020-01-01 00:00:00 my_str_basket[my_key]:2.0",
    )
    assert expected in caplog.record_tuples


def test_LogChannels_no_tick_dict_basket(caplog):
    logger = LogChannels(log_name="gateway_modules_test_logger")
    gateway = Gateway(
        modules=[
            MyNoTickDictBasket(my_data=csp.const(2.0)),
            logger,
        ],
        channels=MySmallGatewayChannels(),
    )
    with caplog.at_level(logging.INFO, logger="gateway_modules_test_logger"):
        gateway.start(starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=1))
    for logger, _level, _message in caplog.record_tuples:
        assert logger != "gateway_modules_test_logger"


def test_LogChannels_state(caplog):
    logger = LogChannels(
        log_name="gateway_modules_test_logger",
        log_states=True,
    )
    gateway = MyGateway(modules=[MyExampleModule(), logger], channels=MySmallGatewayChannels())
    with caplog.at_level(logging.INFO):
        csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=3))
    # why do we have use 'caplog.text' here? The record_tuples uses some interpolation
    # and for recording state that makes the actual captured log_calls different from
    # the ones in caplog.record_tuples
    assert "2020-01-01 00:00:02 s_example:[2]" in caplog.text


def test_LogChannels_no_state(caplog):
    logger = LogChannels(log_name="gateway_modules_test_logger")
    gateway = MyGateway(modules=[MyExampleModule(), logger], channels=MySmallGatewayChannels())
    with caplog.at_level(logging.INFO):
        csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=3))
    assert "s_example" not in caplog.text


@pytest.mark.parametrize("by_key", [True, False])
def test_get_set_LogChannels_doesnt_break(by_key, caplog):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    logger = LogChannels(log_name="gateway_modules_test_logger")
    gateway = MyGateway(modules=[getter, setter, logger], channels=MyGatewayChannels())
    with caplog.at_level(logging.INFO, logger="gateway_modules_test_logger"):
        out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


def test_LogChannels_nothing_set(caplog):
    logger = LogChannels(log_name="gateway_modules_test_logger")
    gateway = Gateway(
        modules=[
            logger,
        ],
        channels=MySmallGatewayChannels(),
    )
    with caplog.at_level(logging.INFO, logger="gateway_modules_test_logger"):
        gateway.start(starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=1))
    for logger, _level, _message in caplog.record_tuples:
        assert logger != "gateway_modules_test_logger"



import json

import pytest
import starlette.responses

from csp_gateway.server.demo import ExampleData
from csp_gateway.server.web import prepare_response


def test_prepare_response():
    data = ExampleData()
    response = prepare_response(res=data, is_list_model=False, is_dict_basket=False, wrap_in_response=True)
    assert isinstance(response, starlette.responses.Response)
    response = prepare_response(res=data, is_list_model=False, is_dict_basket=False, wrap_in_response=False)
    assert json.loads(response) == [json.loads(data.type_adapter().dump_json(data))]


def test_prepare_response_list():
    data = ExampleData()
    response = prepare_response(res=[data], is_list_model=False, is_dict_basket=False, wrap_in_response=False)
    assert json.loads(response) == [json.loads(data.type_adapter().dump_json(data))]
    response = prepare_response(res=[data], is_list_model=True, is_dict_basket=False, wrap_in_response=False)
    assert json.loads(response) == [json.loads(data.type_adapter().dump_json(data))]
    response = prepare_response(res=(data,), is_list_model=True, is_dict_basket=False, wrap_in_response=False)
    assert json.loads(response) == [json.loads(data.type_adapter().dump_json(data))]


def test_prepare_response_dict():
    data = ExampleData()
    with pytest.raises(AttributeError):
        prepare_response(res={"foo": data}, is_list_model=False, is_dict_basket=False, wrap_in_response=False)

    response = prepare_response(res={"foo": data}, is_list_model=False, is_dict_basket=True, wrap_in_response=False)
    assert json.loads(response) == [json.loads(data.type_adapter().dump_json(data))]



from datetime import date, datetime
from typing import Dict, List, Optional, get_type_hints

import csp
import numpy as np
import pytest
from csp import Enum, Struct, ts
from csp.typing import Numpy1DArray
from pydantic import ValidationError

from csp_gateway import GatewayChannels, GatewayStruct as _Base, State
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD


class Symbology(Enum):
    ID = Enum.auto()


class SimpleOrder(_Base):
    _private: str
    symbol: str
    symbology: Symbology = Symbology.ID
    price: float
    quantity: int
    filled: float = 0  # Note that default is an int
    settlement_date: date
    notes: str = ""
    algo_params: dict


class SimpleOrderChild(SimpleOrder):
    new_field: float


class MyCompositeStruct(_Base):
    order: SimpleOrder
    order_list: List[SimpleOrder]


class MyEnum(Enum):
    ONE = 1
    TWO = 2


class MyCspStruct(Struct):
    a: int
    csp_np_array: Numpy1DArray[float] = np.array([13.0]).view(Numpy1DArray)


class MyStruct(_Base):
    foo: float
    fav_num: int = 68
    my_np_array: Numpy1DArray[float] = np.array([6.0]).view(Numpy1DArray)
    my_csp_struct: MyCspStruct = MyCspStruct(a=9)


class MyGatewayChannels(GatewayChannels):
    my_static: float = 0.0
    my_static_dict: Dict[str, float] = {}
    my_static_list: List[str] = []
    my_channel: ts[MyStruct] = None
    s_my_channel: ts[State[MyStruct]] = None
    my_list_channel: ts[List[MyStruct]] = None
    s_my_list_channel: ts[State[MyStruct]] = None
    my_enum_basket: Dict[MyEnum, ts[MyStruct]] = None
    my_str_basket: Dict[str, ts[MyStruct]] = None
    my_enum_basket_list: Dict[MyEnum, ts[List[MyStruct]]] = None
    my_array_channel: ts[Numpy1DArray[float]] = None


class DerivedChannels(MyGatewayChannels):
    pass


class DerivedDerivedChannels(DerivedChannels):
    pass


class DerivedDerivedDerivedChannels(DerivedDerivedChannels):
    pass


def test_construction():
    channels = MyGatewayChannels()
    assert MyGatewayChannels.my_static == "my_static"
    assert channels.my_static == 0.0
    assert MyGatewayChannels.my_channel == "my_channel"
    assert channels.my_channel is None
    channels.my_static = 1.0
    channels.my_channel = csp.null_ts(MyStruct)


def test_construction_derived():
    # Seems innocuous, but has broken before
    channelss = [
        DerivedChannels(),
        DerivedDerivedChannels(),
        DerivedDerivedDerivedChannels(),
    ]
    for channels in channelss:
        assert MyGatewayChannels.my_static == "my_static"
        assert channels.my_static == 0.0
        assert MyGatewayChannels.my_channel == "my_channel"
        assert channels.my_channel is None
        channels.my_static = 1.0
        channels.my_channel = csp.null_ts(MyStruct)


def test_snapshot_model_type_hints():
    snapshot_model = MyGatewayChannels._snapshot_model
    type_hints = get_type_hints(snapshot_model)
    assert len(type_hints) == 6
    assert type_hints["my_channel"] == Optional[MyStruct]
    assert type_hints["my_list_channel"] == Optional[List[MyStruct]]
    assert type_hints["my_enum_basket"] == Optional[Dict[MyEnum, MyStruct]]
    assert type_hints["my_str_basket"] == Optional[Dict[str, MyStruct]]
    assert type_hints["my_enum_basket_list"] == Optional[Dict[MyEnum, List[MyStruct]]]
    assert type_hints[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD] == Optional[datetime]


def test_snapshot_model_instantiation():
    snapshot_model = MyGatewayChannels._snapshot_model
    pass_through = {
        "my_channel": MyStruct(foo=3.0),
        "my_list_channel": [
            MyStruct(foo=3.0),
            MyStruct(foo=2.0),
        ],
        "my_enum_basket": {MyEnum.ONE: MyStruct(foo=3.0)},
    }
    new_model = snapshot_model(**pass_through)
    assert np.array_equal(new_model.my_channel.my_np_array, np.array([6.0]))
    assert new_model.my_channel.foo == 3.0
    assert new_model.my_list_channel[1].foo == 2.0
    assert new_model.my_str_basket is None


def test_snapshot_model_parse_obj():
    snapshot_model = MyGatewayChannels._snapshot_model
    pass_through = {
        "my_channel": {"foo": 3.0},
        "my_list_channel": [
            {"foo": 3.0},
            {
                "foo": 2.0,
                "fav_num": 70,
                "my_np_array": np.array([5.0, 7.0]).view(Numpy1DArray),
            },
        ],
        "my_enum_basket": {MyEnum.ONE: {"foo": 3.0}},
        "my_str_basket": {"my_key": {"foo": 9.0}, "my_key2": {"foo": 11.3}},
        "my_enum_basket_list": {
            MyEnum.TWO: [{"foo": 3.0}, {"foo": 6.89}],
            MyEnum.ONE: [{"foo": 0.0}],
        },
    }
    new_model = snapshot_model.model_validate(pass_through)
    assert isinstance(new_model.my_channel, MyStruct)
    assert new_model.my_channel.foo == 3.0
    assert new_model.my_list_channel[0].foo == 3.0
    assert new_model.my_list_channel[1].fav_num == 70
    assert np.array_equal(new_model.my_list_channel[1].my_np_array, np.array([5.0, 7.0]))
    assert new_model.my_str_basket["my_key"].foo == 9.0
    assert new_model.my_str_basket["my_key2"].fav_num == 68
    assert len(new_model.my_enum_basket_list[MyEnum.TWO]) == 2


def test_snapshot_fails_on_improper_dict():
    snapshot_model = MyGatewayChannels._snapshot_model
    pass_nonexistant_attribute = {"my_channel": {"bad_attr": 68}}
    with pytest.raises(ValidationError):
        snapshot_model.model_validate(pass_nonexistant_attribute)
    with pytest.raises(ValidationError):
        snapshot_model.model_validate({"missing_channel": {"random": 11}})

    pass_non_list_to_list_channel = {"my_list_channel": {"foo": 3.0}}
    with pytest.raises(ValidationError):
        snapshot_model.model_validate(pass_non_list_to_list_channel)



from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Union, get_args, get_origin

import csp
import orjson
from ccflow import BaseModel
from csp import ts
from csp.adapters.kafka import (
    DateTimeType,
    JSONTextMessageMapper,
    KafkaStartOffset,
    RawTextMessageMapper,
)
from csp.adapters.status import Status
from csp.impl.types.container_type_normalizer import ContainerTypeNormalizer
from deprecation import deprecated
from pydantic import Field, FilePath, PrivateAttr, field_validator, model_validator
from typing_extensions import override

from csp_gateway.server import (
    ChannelSelection,
    EncodedEngineCycle,
    GatewayChannels,
    GatewayModule,
)
from csp_gateway.server.shared.engine_replay import EngineReplay
from csp_gateway.server.shared.json_converter import _convert_orjson_compatible

from .utils import KafkaChannelProcessor

__all__ = (
    "ReplayEngineKafka",
    "KafkaConfiguration",
    "KafkaStartOffset",
    "ReadWriteKafka",
)


# NOTE: If the authorization parameters are not correct, the graph may hang indefinitely
class KafkaConfiguration(BaseModel):
    """
    For more information regarding Kafka, see here:
    https://kafka.apache.org/documentation/#introduction
    """

    broker: str = Field(description="broker URL")
    start_offset: Optional[Union[KafkaStartOffset, timedelta, datetime]] = Field(
        None,
        description="""Signifies where to start the stream playback from (defaults to KafkaStartOffset.LATEST ). Can be
                             one of the KafkaStartOffset enum types,
                             datetime - to replay from the given absolute time
                             timedelta - this will be taken as an absolute offset from starttime to playback from""",
    )

    @field_validator("start_offset", mode="before")
    @classmethod
    def _coerce_start_offset(cls, value):
        # Coerce
        if isinstance(value, str) and value in KafkaStartOffset.__members__:
            return KafkaStartOffset[value]
        return value

    debug: bool = Field(
        False,
        description="Whether to be in debug mode. If True, start_offset is set to None",
    )
    group_id: Optional[str] = Field(
        None,
        description="""
        If set, will cause the adapter to behave as a consume-once consumer.
        start_offset may not be set in this case since the adapter will
        always replay from the last consumed offset.""",
    )
    group_id_prefix: str = Field(
        "",
        description="""When not passing an explicit group_id, a prefix can be supplied that will be
        used to prefix the UUID generated for the group_id""",
    )
    max_threads: int = 4
    max_queue_size: int = 1000000
    poll_timeout: timedelta = Field(
        timedelta(seconds=1),
        description="""
        Amount of time to wait if data is not available in the broker
        to consume.""",
    )

    auth: bool = Field(False, description="Determines whether to use authentication")
    security_protocol: str = Field("SASL_SSL", description="Security protocol, only used if auth is set to True")
    sasl_kerberos_keytab: Optional[FilePath] = Field(
        None,
        description="""
        Location of the Kerberos keytab for Kerberos authentication.
        Only used if auth is set to True
        """,
    )
    sasl_kerberos_principal: str = Field("", description="Name of kerberos principal, only used if auth is set to True")
    ssl_ca_location: Optional[FilePath] = Field(
        "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",
        description="""
        Location of the PEM key for broker validation.
        Only used if auth is set to True. Defaults to a file location many linux machines
        have access to.""",
    )
    sasl_kerberos_service_name: str = Field("kafka", description="Kerberos service name, only used if auth is set to True")

    rd_kafka_conf_options: Optional[Dict] = Field(
        None,
        description="""
        Extra configuration options that will be directly passed to the C++ Kafka consumers and producers.
        For example, if you want to change maximum amount of time the Kafka broker will block before
        answering the fetch request if there isn't sufficient data to immediately satisfy the requirement
        given by fetch.min.bytes, you can set:
        rd_kafka_conf_options={'fetch.wait.max.ms' : '10' }
        """,
    )


# TODO: Does not work with dict baskets
class ReadWriteKafka(GatewayModule):
    """
    This class is designed to:
    1. Publish Gateway channel ticks to Kafka.
    2. Read Gateway channel ticks from Kafka to instantiate
       those ticks on the corresponding channel.

    Custom serialization and deserialization logic can be provided by
    inheriting from this class and overriding the following function:
    * `serialize`
    * `deserialize`

    However, since the defaults provide helpful functionality, other functions can be
    overriden in order to still utilize some of this default behavior.

    * `serialize_to_python` transforms an object into a form compatible with being written to json.

    *`deserialize_to_python` takes a parsed json object, and an expected type, and can then transform the object before being parsed to the appropriate type

    *`deserialize_to_target` takes a python object, and transforms it to the target type

    Caveats:
    - Dict baskets are not supported.
    """

    config: KafkaConfiguration
    requires: Optional[ChannelSelection] = []
    publish_channel_to_topic_and_key: Dict[str, Dict[str, str]] = {}
    subscribe_channel_to_topic_and_key: Dict[str, Dict[str, str]] = {}

    publish_channel_processors: Dict[str, KafkaChannelProcessor] = Field(
        default={},
        description=(
            "Dictionary mapping channel to KafkaChannelProcessor that "
            "performs a transformation on the incoming tick before publishing "
            "to Kafka. "
            "If the procesing function returns None for a tick, that tick is not published to Kafka"
        ),
    )
    subscribe_channel_processors: Dict[str, KafkaChannelProcessor] = Field(
        default={},
        description=(
            "Dictionary mapping channel to KafkaChannelProcessor processor that "
            "performs a transformation on the incoming tick before pushing the "
            "tick to the gateway. "
            "If the procesing function returns None for a tick, that tick is not pushed into the corresponding channel."
        ),
    )

    encoding_with_engine_timestamps: bool = Field(
        default=False,
        description=("Determines whether the encodings to and from Kafka include the csp engine timestamp."),
    )
    subscribe_with_csp_engine_timestamp: bool = Field(
        default=False,
        description=(
            "Determines whether the timestamp used when subscribing is the "
            "timestamp Kafka received the message, or the timestamp included on the encoding."
        ),
    )
    subscribe_with_struct_id: bool = Field(
        default=False,
        description=("If False, replaces the id field on GatewayStructs from Kafka with one autogenerated by the current Gateway."),
    )
    subscribe_with_struct_timestamp: bool = Field(
        default=False,
        description=("If False, replaces the timestamp field on the GatewayStruct with a timestamp autogenerated by the current Gateway."),
    )
    include_subscribe_messages_before_engine_start: bool = Field(
        default=False,
        description="Based on the Kafka offset, if Kafka messages are read in with timestamps before engine start time (either the Kafka timestamp or the message on the struct), whether these messages will get dropped or included in the graph. The default is False, only pulling in messages into the csp graph that are at or after the engine starttime. If True, these messages will be pulled in with timestamp equal to engine starttime.",
    )
    _kafkaadapter: csp.adapters.kafka.KafkaAdapterManager = PrivateAttr(default=None)

    @model_validator(mode="after")
    def check_parameters_consistent(self):
        if self.subscribe_with_csp_engine_timestamp and not self.encoding_with_engine_timestamps:
            raise ValueError("If `subscribe_with_csp_engine_timestamp` is True, you must set `encoding_with_engine_timestamp` to be True as well")
        return self

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._kafkaadapter = csp.adapters.kafka.KafkaAdapterManager(
            broker=self.config.broker,
            start_offset=self.config.start_offset,
            group_id=self.config.group_id,
            group_id_prefix=self.config.group_id_prefix,
            max_threads=self.config.max_threads,
            max_queue_size=self.config.max_queue_size,
            auth=self.config.auth,
            security_protocol=self.config.security_protocol,
            sasl_kerberos_keytab=str(self.config.sasl_kerberos_keytab),
            sasl_kerberos_principal=self.config.sasl_kerberos_principal,
            ssl_ca_location=str(self.config.ssl_ca_location),
            sasl_kerberos_service_name=self.config.sasl_kerberos_service_name,
            rd_kafka_conf_options=self.config.rd_kafka_conf_options,
            debug=self.config.debug,
            poll_timeout=self.config.poll_timeout,
        )

    def deserialize_to_python(self, obj: object, ts_typ: object) -> object:
        """Intercepts encoding from Kafka before conversion to to target type.
        This can be overwritten to implement custom logic to alter the object before it is passed onto the `deserialize_to_target` to create the csp struct
        """
        return obj

    def deserialize_to_target(self, obj: object, ts_typ: object) -> object:
        """Receives a python object loaded from json, and converts it to an object of type `ts_type`"""
        normalized_type = ContainerTypeNormalizer.normalize_type(ts_typ)
        is_list = get_origin(normalized_type) is list

        context = {}
        if not self.subscribe_with_struct_id:
            context["force_new_id"] = True
        if not self.subscribe_with_struct_timestamp:
            context["force_new_timestamp"] = True
        if is_list:
            inner_type = get_args(normalized_type)[0]
            type_adapter = inner_type.type_adapter()
            return [type_adapter.validate_python(v, context=context) for v in obj]
        return ts_typ.type_adapter().validate_python(obj, context=context)

    def deserialize(self, encoding: str, ts_typ: object):
        """Receives the encoded string from Kafka and
        deserializes it appropriately to convert into an object
        of type specified by `ts_typ`.
        """
        encoding_obj = orjson.loads(encoding)

        json_dict = self.deserialize_to_python(encoding_obj, ts_typ=ts_typ)

        return self.deserialize_to_target(json_dict, ts_typ=ts_typ)

    @deprecated(details="Use serialize_to_python instead.")
    def serialize_to_dict(self, x: Any) -> Dict[str, Any]:
        """Serializes an object to a dictionary compatible with
        orjson for encoding to json."""
        return self.serialize_to_python(x)

    def serialize(self, x: Any) -> str:
        """Serializes an object to json"""
        encoding_for_orjson = self.serialize_to_python(x)
        return orjson.dumps(
            encoding_for_orjson,
            option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_NAIVE_UTC,
        ).decode()

    def serialize_to_python(self, obj: Any) -> Any:
        """Serializes an object to a format compatible with being transformed into json"""
        return _convert_orjson_compatible(obj)

    @csp.node
    def deserialize_csp(self, encoding: ts[str], ts_typ: "T") -> ts["T"]:  # noqa
        if csp.ticked(encoding):
            return self.deserialize(encoding, ts_typ)

    @csp.node
    def serialize_csp(self, x: ts[object]) -> ts[str]:
        return self.serialize(x)

    @csp.node
    def serialize_with_engine_timestamp_csp(self, x: ts[object]) -> ts[EncodedEngineCycle]:
        encoding = self.serialize(x)
        return EncodedEngineCycle(encoding=encoding, csp_timestamp=csp.now())

    def connect(self, channels: GatewayChannels):
        if self.encoding_with_engine_timestamps:
            msg_mapper = JSONTextMessageMapper(datetime_type=DateTimeType.UINT64_MILLIS)
            publish_field_map = {
                "encoding": "encoding",
                "csp_timestamp": "csp_timestamp",
            }
            subscribe_field_map = publish_field_map
        else:
            msg_mapper = RawTextMessageMapper()
            publish_field_map = None
            subscribe_field_map = {"": "encoding"}

        for channel_name, topic_to_key in self.publish_channel_to_topic_and_key.items():
            channel = channels.get_channel(channel_name)
            channel_type = channels.get_outer_type(channel_name).typ
            channel_processor = self.publish_channel_processors.get(channel_name)

            for topic, key in topic_to_key.items():
                if channel_processor is not None:
                    raw_value = channel_processor.apply_process(channel_type, channel, topic, key)
                else:
                    raw_value = channel

                if self.encoding_with_engine_timestamps:
                    encoded_value = self.serialize_with_engine_timestamp_csp(raw_value)
                else:
                    encoded_value = self.serialize_csp(raw_value)
                self._kafkaadapter.publish(
                    msg_mapper=msg_mapper,
                    field_map=publish_field_map,
                    topic=topic,
                    key=key,
                    x=encoded_value,
                )

        tick_timestamp_from_field = "csp_timestamp" if self.subscribe_with_csp_engine_timestamp else None

        for (
            channel_name,
            topic_to_key,
        ) in self.subscribe_channel_to_topic_and_key.items():
            values_to_tick = []
            for topic, key in topic_to_key.items():
                sub = self._kafkaadapter.subscribe(
                    ts_type=EncodedEngineCycle,
                    topic=topic,
                    msg_mapper=msg_mapper,
                    key=key,
                    field_map=subscribe_field_map,
                    tick_timestamp_from_field=tick_timestamp_from_field,
                    adjust_out_of_order_time=True,
                    include_msg_before_start_time=self.include_subscribe_messages_before_engine_start,
                    push_mode=csp.PushMode.NON_COLLAPSING,
                ).encoding
                channel_type = channels.get_outer_type(channel_name).typ
                deserialized_sub = self.deserialize_csp(encoding=sub, ts_typ=channel_type)
                channel_processor = self.subscribe_channel_processors.get(channel_name)

                if channel_processor is not None:
                    val_to_tick = channel_processor.apply_process(channel_type, deserialized_sub, topic, key)
                else:
                    val_to_tick = deserialized_sub
                values_to_tick.append(val_to_tick)

            channels.set_channel(
                channel_name,
                csp.flatten(values_to_tick),
            )

    def status(self) -> ts[Status]:
        """Returns a ticking edge of the Status updates from the underlying csp Kafka adapter."""
        return self._kafkaadapter.status()


class ReplayEngineKafka(EngineReplay):
    key: str = Field(
        description="""
        The key within a topic. Events published on the same topic with the same key go to the same partition.
        Kafka guarantees that any consumer of a given topic-partition will always read that
        partition's events in exactly the same order as they were written."""
    )
    topic: str = Field(description="Kafka topic")
    config: KafkaConfiguration
    _kafkaadapter: Any = PrivateAttr(default=None)

    # backwards compatibility, if the config is set make sure nothing from it is
    @model_validator(mode="before")
    @classmethod
    def validate_kafka(cls, values):
        config_fields = [
            "broker",
            "start_offset",
            "debug",
            "group_id",
            "group_id_prefix",
            "max_threads",
            "max_queue_size",
            "poll_timeout",
            "auth",
            "security_protocol",
            "sasl_kerberos_keytab",
            "sasl_kerberos_principal",
            "ssl_ca_location",
            "sasl_kerberos_service_name",
            "rd_kafka_conf_options",
        ]
        if "config" in values:
            if duplicate_values := [name for name in config_fields if name in values]:
                raise ValueError(f"Config is set, these attributes should be on the Config: {duplicate_values}")
        else:
            config_args = {}
            # we set the values on the config
            for field in config_fields:
                if field in values:
                    config_args[field] = values.pop(field)
            values["config"] = config_args
        return values

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._kafkaadapter = csp.adapters.kafka.KafkaAdapterManager(
            broker=self.config.broker,
            start_offset=self.config.start_offset,
            group_id=self.config.group_id,
            group_id_prefix=self.config.group_id_prefix,
            max_threads=self.config.max_threads,
            max_queue_size=self.config.max_queue_size,
            auth=self.config.auth,
            security_protocol=self.config.security_protocol,
            sasl_kerberos_keytab=str(self.config.sasl_kerberos_keytab),
            sasl_kerberos_principal=self.config.sasl_kerberos_principal,
            ssl_ca_location=str(self.config.ssl_ca_location),
            sasl_kerberos_service_name=self.config.sasl_kerberos_service_name,
            rd_kafka_conf_options=self.config.rd_kafka_conf_options,
            debug=self.config.debug,
            poll_timeout=self.config.poll_timeout,
        )

    @override
    def subscribe(self):
        return self._kafkaadapter.subscribe(
            ts_type=EncodedEngineCycle,
            msg_mapper=JSONTextMessageMapper(datetime_type=DateTimeType.UINT64_MILLIS),
            topic=self.topic,
            key=self.key,
            field_map={"encoding": "encoding", "csp_timestamp": "csp_timestamp"},
            meta_field_map={"timestamp": "csp_timestamp"},  # don't use the timestamp provided by kafka, but the one encoded on the message
            # adjust_out_of_order_time=True,  we do this in python and log when it happens
            push_mode=csp.PushMode.NON_COLLAPSING,
        ).encoding

    @override
    def publish(self, encoded_channels: ts[EncodedEngineCycle]):
        self._kafkaadapter.publish(
            msg_mapper=JSONTextMessageMapper(datetime_type=DateTimeType.UINT64_MILLIS),
            field_map={"encoding": "encoding", "csp_timestamp": "csp_timestamp"},
            topic=self.topic,
            key=self.key,
            x=encoded_channels,
        )

    @override
    def status(self) -> ts[Status]:
        """Returns a ticking edge of the Status updates from the underlying csp Kafka adapter."""
        return self._kafkaadapter.status()



from typing import List, Optional, Set, Union, get_args, get_origin

from fastapi import APIRouter, HTTPException, Request
from pydantic import BaseModel

from csp_gateway.server import ChannelSelection

from ..utils import get_default_responses
from .shared import prepare_response

__all__ = (
    "add_lookup_routes",
    "add_lookup_available_channels",
)


def add_lookup_routes(
    api_router: APIRouter,
    field: str,
    model: Union[BaseModel, List[BaseModel]],
) -> None:
    if model and get_origin(model) is list:
        model = get_args(model)[0]

    async def lookup(id: str, request: Request) -> List[model]:  # type: ignore[misc, valid-type]
        """
        This endpoint lets you lookup any GatewayStruct by its uniquely generated `id`.
        """
        # Throw 404 if not a supported channel
        if not hasattr(request.app.gateway.channels, field):
            raise HTTPException(status_code=404, detail="Channel not found: {}".format(field))

        # lookup by id
        res = model.lookup(id)

        return prepare_response(res, is_list_model=False)

    api_router.get(
        "/{}/{{id:path}}".format(field),
        responses=get_default_responses(),
        response_model=List[model],
        name="Lookup {}".format(field),
    )(lookup)

    api_router.get(
        "/{}/{{id:path}}".format(field.replace("_", "-")),
        responses=get_default_responses(),
        response_model=List[model],
        include_in_schema=False,
    )(lookup)


def add_lookup_available_channels(api_router: APIRouter, fields: Optional[Set[str]] = None) -> None:
    @api_router.get(
        "/",
        responses=get_default_responses(),
        response_model=List[str],
    )
    async def get_lookup(request: Request) -> List[str]:
        """
        This endpoint will return a list of string values of all available channels under the `/lookup` route.
        """
        return sorted(ChannelSelection().select_from(request.app.gateway.channels) if fields is None else fields)



import os

import hydra

import csp_gateway.server.config
from csp_gateway import Gateway


def test_config_file_load():
    """This case tests a config where the gateway is specified in the user config file"""
    user_config_dir = os.path.dirname(__file__)
    r = csp_gateway.server.config.load_config(
        overwrite=True,
        config_dir=user_config_dir,
        overrides=["+user_config=sample_config"],
    )
    try:
        assert isinstance(r["gateway"], Gateway)
    finally:
        r.clear()


def test_config_file_load_gateway():
    user_config_dir = os.path.dirname(__file__)
    g = csp_gateway.server.config.load_gateway(
        overwrite=True,
        config_dir=user_config_dir,
        overrides=["+user_config=sample_config"],
    )
    assert isinstance(g, Gateway)


def test_start_load():
    config_dir = os.path.join(os.path.dirname(__file__), "../../server/config")
    with hydra.initialize_config_dir(version_base=None, config_dir=config_dir):
        cfg = hydra.compose(config_name="base.yaml")
    assert dict(cfg["start"]) == {
        "realtime": True,
        "block": False,
        "show": False,
        "rest": True,
        "ui": True,
    }



import threading

import orjson
from typing_extensions import override

from ..struct import GatewayStruct


class Controls(GatewayStruct):
    name: str = "none"
    status: str = "none"
    data: dict = {}
    data_str: str = ""

    # TODO: all `GatewayStructs`` cross thread boundaries and should
    # have locks, but just doing `Controls`` for now
    _lock: object

    def lock(self, blocking=True, timeout=-1):
        # FIXME: first call is...not threadsafe
        if not hasattr(self, "_lock"):
            self._lock = threading.Lock()
        self._lock.acquire(blocking=blocking, timeout=timeout)

    def unlock(self):
        if not hasattr(self, "_lock"):
            self._lock = threading.Lock()
        if self._lock.locked():
            self._lock.release()

    def update_str(self):
        #  NOTE: This is needed to resolve a race condition between
        #  Web API and Perspective. We need to create a data_str
        #  because perspective cannot read python dict types and
        #  csp.Structs.to_json creates dictionaries. This will
        #  be removed when better to_json support is added.
        self.lock()
        if self.data_str == "":
            self.data_str = orjson.dumps(self.data).decode()
        self.unlock()

    @override
    def psp_flatten(self, custom_jsonifier=None):
        self.update_str()
        res = super().psp_flatten(custom_jsonifier)
        return res



from datetime import datetime, timedelta
from inspect import getframeinfo, stack
from pprint import pprint
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    get_origin,
)

import csp
import numpy as np
import pandas as pd
from csp import ts
from pydantic import BaseModel, ConfigDict, Field

from csp_gateway import ChannelSelection, GatewayChannels, GatewayModule

__all__ = ("GatewayTestHarness",)

ChannelType = Union[str, Tuple[str, Any]]
T = TypeVar("T")


def assert_equal(actual, desired):
    """Implement custom equality assertion function"""
    # Right now we leverage numpy's, which is pretty general in that it handles standard objects,
    # nan's, and numpy arrays (containing nans)
    # Some things it might not handle properly are pd.NaT and pd.NA
    np.testing.assert_equal(actual, desired)


def assert_not_equal(actual, desired):
    # We do not use something like np.testing.assert_raises(AssertionError, assert_equal, actual, desired))
    # because when it fails, we want the error message in the assertion to show both actual/desired
    # As a result, this function won't handle numpy arrays well.
    if pd.isna(desired):
        assert pd.isna(actual)
    else:
        assert actual != desired


def assert_almost_equal(actual, desired):
    np.testing.assert_almost_equal(actual, desired)


class BaseGatewayTestEvent(BaseModel):
    """Base class for events that happen during the test.

    These events should not be constructed directly. Rather, they are constructed from the GatewayTestHarness
    """

    # Traceback info to help with debugging
    _lineno: int = None
    _filename: str = None

    model_config = ConfigDict(frozen=True, extra="forbid")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Try to go two levels up in the stack, from where the helper functions on
        # GatewayTestHarness are themselves called.
        try:
            caller = getframeinfo(stack()[2][0])
            self._lineno = caller.lineno
            self._filename = caller.filename
        except Exception:
            pass

    def apply(self, now, values, tick_counts, ticked_values, *args, **kwargs) -> Optional[bool]:
        """Function to run inside the csp graph during runtime.
        If the function returns True, it will reset the state of ticked values and counts.
        """
        ...


class GatewayResetEvent(BaseGatewayTestEvent):
    """Event to explicitly reset the ticked counts and values."""

    def apply(self, *args, **kwargs):
        return True


class GatewayDelayEvent(BaseGatewayTestEvent):
    delay: Union[timedelta, datetime] = Field(
        timedelta(seconds=1), description="Time to wait before applying the event or the time to apply the event."
    )


class GatewayDataEvent(BaseGatewayTestEvent):
    """Event that sends data into the gateway"""

    channel: ChannelType
    value: Any  # GatewayStruct


class GatewayPrintEvent(BaseGatewayTestEvent):
    msg: str

    def apply(self, now, *args, **kwargs):
        print(f"{now}: {self.msg}")


class GatewayPrintTickedEvent(BaseGatewayTestEvent):
    def apply(self, now, values, *args, **kwargs):
        print(f"{now}: Ticked values:")
        pprint(values)


class GatewayPrintTickCountsEvent(BaseGatewayTestEvent):
    def apply(self, now, values, tick_counts, *args, **kwargs):
        print(f"{now}: Tick counts:")
        pprint(tick_counts)


class GatewayAssertTickCountsEqualEvent(BaseGatewayTestEvent):
    tick_counts: Dict[ChannelType, int]

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert tick_counts == self.tick_counts


class GatewayAssertTickCountEvent(BaseGatewayTestEvent):
    channel: ChannelType
    count: int

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert tick_counts.get(self.channel, 0) == self.count


class GatewayAssertEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    value: Any
    almost: bool = False

    def apply(self, now, values, tick_counts, *args, **kwargs):
        if self.almost:
            assert_almost_equal(values[self.channel], self.value)
        else:
            assert_equal(values[self.channel], self.value)


class GatewayAssertTypeEvent(BaseGatewayTestEvent):
    channel: ChannelType
    value: Any

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert isinstance(values[self.channel], self.value)


class GatewayAssertAttrEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    value: Any
    attr: str
    almost: bool = False

    def apply(self, now, values, tick_counts, *args, **kwargs):
        if self.almost:
            assert_almost_equal(getattr(values[self.channel], self.attr), self.value)
        else:
            assert_equal(getattr(values[self.channel], self.attr), self.value)


class GatewayAssertAttrsEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    values: Dict[str, Any]
    almost: bool = False

    def apply(self, now, values, tick_counts, *args, **kwargs):
        for attr, value in self.values.items():
            if self.almost:
                assert_almost_equal(getattr(values[self.channel], attr), value)
            else:
                assert_equal(getattr(values[self.channel], attr), value)


class GatewayAssertAttrNotEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    value: Any
    attr: str

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert_not_equal(getattr(values[self.channel], self.attr), self.value)


class GatewayAssertAttrUnsetEvent(BaseGatewayTestEvent):
    channel: ChannelType
    unset: bool
    attr: str

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert hasattr(values[self.channel], self.attr) == (not self.unset)


class GatewayAssertLenEvent(BaseGatewayTestEvent):
    channel: ChannelType
    value: int

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert len(values[self.channel]) == self.value


class GatewayAssertIdxTypeEvent(BaseGatewayTestEvent):
    channel: ChannelType
    idx: int
    value: Any

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert isinstance(values[self.channel][self.idx], self.value)


class GatewayAssertIdxAttrEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    idx: int
    value: Any
    attr: str
    almost: bool = False

    def apply(self, now, values, tick_counts, *args, **kwargs):
        if self.almost:
            assert_almost_equal(getattr(values[self.channel][self.idx], self.attr), self.value)
        else:
            assert_equal(getattr(values[self.channel][self.idx], self.attr), self.value)


class GatewayAssertIdxAttrsEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    idx: int
    values: Dict[str, Any]
    almost: bool = False

    def apply(self, now, values, tick_counts, *args, **kwargs):
        for attr, value in self.values.items():
            if self.almost:
                assert_almost_equal(getattr(values[self.channel][self.idx], attr), value)
            else:
                assert_equal(getattr(values[self.channel][self.idx], attr), value)


class GatewayAssertIdxAttrNotEqualEvent(BaseGatewayTestEvent):
    channel: ChannelType
    idx: int
    value: Any
    attr: str

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert_not_equal(getattr(values[self.channel][self.idx], self.attr), self.value)


class GatewayAssertIdxAttrUnsetEvent(BaseGatewayTestEvent):
    channel: ChannelType
    idx: int
    unset: bool
    attr: str

    def apply(self, now, values, tick_counts, *args, **kwargs):
        assert hasattr(values[self.channel][self.idx], self.attr) == (not self.unset)


class GatewayAssertTickedEvents(BaseGatewayTestEvent):
    channel: ChannelType
    assert_func: Callable[[Sequence[Tuple[datetime, Any]]], None]

    def apply(self, now, values, tick_counts, ticked_values, *args, **kwargs):
        self.assert_func(ticked_values[self.channel])


class GatewayEvaluateCallableEvent(BaseGatewayTestEvent):
    f: Callable[[], None]

    def apply(self, now, values, tick_counts, ticked_values, *args, **kwargs) -> Optional[bool]:
        self.f()
        return None


@csp.node
def _curve(data: List[Tuple[Union[datetime, timedelta], "T"]]) -> ts["T"]:
    """TODO: csp.curve doesn't allow mixing datetime and timedelta at the moment. Switch to csp.curve once supported."""
    with csp.alarms():
        alarm = csp.alarm(object)

    with csp.start():
        for t, datum in data:
            csp.schedule_alarm(alarm, t, datum)

    if csp.ticked(alarm):
        return alarm


class GatewayTestHarness(GatewayModule):
    name: str = "GatewayTestHarness"
    test_channels: ChannelSelection = Field(description="List of channels to test (both inputs and outputs)")
    events: List[BaseGatewayTestEvent] = []
    verbose: bool = False
    test_dynamic_keys: Dict[str, List[str]] = {}

    is_performance_test: bool = Field(
        default=False,
        description=(
            "Whether the test harness is running in a performance test. Normally outputs are captured so that "
            "assertions can be done on them, but this has some overhead. Turning performance test mode on causes "
            "the outputs to not be captured. This means that assertions will probably fail."
        ),
    )

    def dynamic_keys(self):
        return self.test_dynamic_keys

    def connect(self, channels: GatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        # This is done statically at graph building time,
        # so that we can segregate the types properly for an arbitrary set of channels
        delta = timedelta(0)
        event_time = timedelta(0)
        reference_time = None
        event_curve_data = []
        test_channels = self.test_channels.select_from(channels=channels, all_fields=True)
        curve_data = {channel: [] for channel in test_channels}
        for event in self.events:
            if isinstance(event, GatewayDelayEvent):
                if isinstance(event.delay, timedelta):
                    delta += event.delay
                else:
                    # event.delay is a datetime.
                    if reference_time is not None:
                        if event_time > event.delay:
                            raise ValueError(f"Trying to move backwards in time from {event_time} to {event.delay}.")

                    reference_time = event.delay
                    delta = timedelta(0)

                if reference_time is None:
                    event_time = delta
                else:
                    event_time = reference_time + delta

            if isinstance(event, GatewayDataEvent):
                curve_data[event.channel].append((event_time, event.value))

            event_curve_data.append((event_time, event))

        channel_values = {}
        for channel in test_channels:
            outer_type = channels.get_outer_type(channel)
            if get_origin(outer_type) is dict:
                basket = channels.get_channel(channel)

                # Split the data events for the basket by basket key.
                basket_curve_data = {k: [] for k in basket.keys()}
                if curve_data[channel]:
                    for delta, basket_events in curve_data[channel]:
                        if not isinstance(basket_events, dict):
                            raise ValueError(f"Event values for dictionary baskets should be a dictionary, got {basket_events}.")

                        for k, v in basket_events.items():
                            if k not in basket:
                                raise ValueError(f"{k} is not a valid key in {channel}.")

                            basket_curve_data[k].append((delta, v))

                # Go through the basket, putting the values into channel_values and create the curves for
                # each basket key.
                curves = {}
                for k, v in basket.items():
                    # Channel output
                    channel_values[(channel, k)] = v
                    if self.verbose:
                        csp.print(f"Update {str((channel, k))}", v)

                    # Channel input
                    typ = v.tstype.typ
                    if basket_curve_data[k]:
                        curves[k] = _curve.using(T=typ)(basket_curve_data[k])
                    else:
                        curves[k] = csp.null_ts(typ)

                # Set input data to channel.
                channels.set_channel(channel, curves)
            else:
                typ = outer_type.typ
                if curve_data[channel]:
                    curve = _curve.using(T=typ)(curve_data[channel])
                else:
                    curve = csp.null_ts(typ)
                channels.set_channel(channel, curve)
                channel_values[channel] = channels.get_channel(channel)
                if self.verbose:
                    csp.print(f"Update {channel} ", channel_values[channel])

        if event_curve_data:
            event_curve = _curve.using(T=BaseGatewayTestEvent)(event_curve_data)
        else:
            event_curve = csp.null_ts(BaseGatewayTestEvent)

        # Run the tests
        self._run_test_events(event_curve, channel_values)

    @csp.node
    def _run_test_events(self, event: ts[BaseGatewayTestEvent], channel_values: Dict[object, ts[object]]):
        with csp.state():
            s_values = {}
            print(f"Starting {self.name}")

        with csp.stop():
            print(f"Finishing {self.name}")

        if csp.ticked(channel_values) and not self.is_performance_test:
            tickeditems = dict(channel_values.tickeditems())
            for channel, value in tickeditems.items():
                if channel not in s_values:
                    s_values[channel] = []

                if isinstance(channel, str) and channel.startswith("s_"):
                    s_values[channel].append((csp.now(), value.query()))
                else:
                    s_values[channel].append((csp.now(), value))

        if csp.ticked(event):
            try:
                current_values = {k: v[-1][1] for k, v in s_values.items() if len(v) > 0}
                tick_counts = {k: len(v) for k, v in s_values.items()}
                reset = event.apply(csp.now(), current_values, tick_counts, ticked_values=s_values)
            except AssertionError as e:
                print(f"Assertion failed in test at {event._filename}:{event._lineno}")
                raise e
            except Exception as e:
                print(f"Exception in test at {event._filename}:{event._lineno}")
                raise e
            if reset:
                s_values = {}

    def advance(self, *, delay: Union[timedelta, datetime] = timedelta(seconds=1), msg: str = "", pre_msg: str = "") -> None:  # noqa: B008
        """Convenience function to reset the state, and advance to the next part of the test by adding a delay.
        Optional messages are printed to help delimit sections of the test.

        Args:
            delay: Time to advance (timedelta) or the time (datetime) to advance to.
            msg: Message to print after the delay.
            pre_msg: Message to print before the delay.
        """
        self.reset()
        self.print(pre_msg)
        self.delay(delay)
        self.print(msg)

    def reset(self):
        self.events.append(GatewayResetEvent())

    def delay(self, delay: Union[timedelta, datetime]):
        """Move forward in time.

        Args:
            delay: Time to move forward by or the time to move forward to.
        """
        self.events.append(GatewayDelayEvent(delay=delay))

    def print(self, msg):
        self.events.append(GatewayPrintEvent(msg=msg))

    def print_ticked(self):
        self.events.append(GatewayPrintTickedEvent())

    def print_tick_counts(self):
        self.events.append(GatewayPrintTickCountsEvent())

    def send(self, channel, value):
        self.events.append(GatewayDataEvent(channel=channel, value=value))

    def assert_tick_counts(self, tick_counts):
        self.events.append(GatewayAssertTickCountsEqualEvent(tick_counts=tick_counts))

    def assert_ticked(self, channel, count=1):
        self.events.append(GatewayAssertTickCountEvent(channel=channel, count=count))

    def assert_equal(self, channel, value):
        self.events.append(GatewayAssertEqualEvent(channel=channel, value=value))

    def assert_almost_equal(self, channel, value):
        self.events.append(GatewayAssertEqualEvent(channel=channel, value=value, almost=True))

    def assert_type(self, channel, value):
        self.events.append(GatewayAssertTypeEvent(channel=channel, value=value))

    def assert_attr_equal(self, channel, attr, value):
        self.events.append(GatewayAssertAttrEqualEvent(channel=channel, attr=attr, value=value))

    def assert_attr_almost_equal(self, channel, attr, value):
        self.events.append(GatewayAssertAttrEqualEvent(channel=channel, attr=attr, value=value, almost=True))

    def assert_attrs_equal(self, channel, values):
        self.events.append(GatewayAssertAttrsEqualEvent(channel=channel, values=values))

    def assert_attrs_almost_equal(self, channel, values):
        self.events.append(GatewayAssertAttrsEqualEvent(channel=channel, values=values, almost=True))

    def assert_attr_not_equal(self, channel, attr, value):
        self.events.append(GatewayAssertAttrNotEqualEvent(channel=channel, attr=attr, value=value))

    def assert_attr_unset(self, channel, attr, unset=True):
        self.events.append(GatewayAssertAttrUnsetEvent(channel=channel, attr=attr, unset=unset))

    def assert_idx_type(self, channel, idx, value):
        self.events.append(GatewayAssertIdxTypeEvent(channel=channel, idx=idx, value=value))

    def assert_idx_attr_equal(self, channel, idx, attr, value):
        self.events.append(GatewayAssertIdxAttrEqualEvent(channel=channel, idx=idx, attr=attr, value=value))

    def assert_idx_attr_almost_equal(self, channel, idx, attr, value):
        self.events.append(GatewayAssertIdxAttrEqualEvent(channel=channel, idx=idx, attr=attr, value=value, almost=True))

    def assert_idx_attrs_equal(self, channel, idx, values):
        self.events.append(GatewayAssertIdxAttrsEqualEvent(channel=channel, idx=idx, values=values))

    def assert_idx_attrs_almost_equal(self, channel, idx, values):
        self.events.append(GatewayAssertIdxAttrsEqualEvent(channel=channel, idx=idx, values=values, almost=True))

    def assert_idx_attr_not_equal(self, channel, idx, attr, value):
        self.events.append(GatewayAssertIdxAttrNotEqualEvent(channel=channel, idx=idx, attr=attr, value=value))

    def assert_idx_attr_unset(self, channel, idx, attr, unset=True):
        self.events.append(GatewayAssertIdxAttrUnsetEvent(channel=channel, idx=idx, attr=attr, unset=unset))

    def assert_len(self, channel, value):
        self.events.append(GatewayAssertLenEvent(channel=channel, value=value))

    def assert_ticked_values(self, channel, assert_func: Callable[[Sequence[Tuple[datetime, Any]]], None]):
        """Apply an assert_func to the ticked values on a particular channel.

        :param channel: Channel to apply the assert_func on.
        :param assert_func: Function that accepts a sequence of ticked values on a channel represented as a sequence
            of tuples where the first item is the engine time of the tick and the second item is the ticked value.
            This function should raise an AssertionError if the ticked values is not what is expected.
        """
        self.events.append(GatewayAssertTickedEvents(channel=channel, assert_func=assert_func))

    def assert_value(self, channel, assert_func: Callable[[Any], None]):
        """Apply an assert_func to the current value on a particular channel.

        :param channel: Channel to apply the assert_func on.
        :param assert_func: Function that takes accepts the current value of a particular channel. This function should
            raise an AssertionError if the value is not what is expected.
        """

        def ticked_values_assert_func(ticked_values):
            assert len(ticked_values) > 0
            assert_func(ticked_values[-1][1])

        self.assert_ticked_values(channel, ticked_values_assert_func)

    def eval(self, f: Callable[[], None]):
        """Evaluate a function whilst the circuit is running. Useful for things like starting and stopping a profiler
        whilst the circuit is running."""
        self.events.append(GatewayEvaluateCallableEvent(f=f))



import asyncio
from asyncio import AbstractEventLoop, Future as AsyncFuture
from concurrent.futures import Future as ConcurrentFuture
from typing import Any, Dict, Union

import csp
from csp import ts
from csp.impl.genericpushadapter import GenericPushAdapter

EitherFutureType = Union[AsyncFuture, ConcurrentFuture]


class FutureAdapter(GenericPushAdapter):
    def __init__(self, kind: str, loop: AbstractEventLoop = None, name: str = None) -> None:
        if kind in ("async", AsyncFuture):
            super().__init__(AsyncFuture, name=name)
            self._kind = "async"
            self._loop = loop or asyncio.get_event_loop()
        else:
            super().__init__(ConcurrentFuture, name=name)
            self._kind = "concurrent"

    def push_tick(self, future: EitherFutureType = None) -> EitherFutureType:
        if self._kind == "async":
            fut = future or AsyncFuture(loop=self._loop)
        elif self._kind == "concurrent":
            fut = future or ConcurrentFuture()
        super().push_tick(fut)
        return fut


def on_request(value: ts[object], future: ts[object]):  # type: ignore[no-untyped-def]
    with csp.start():
        csp.make_passive(value)

    if csp.ticked(future):
        # TODO consider returning something other than None?
        to_return = value if csp.valid(value) else None

        if isinstance(future, AsyncFuture):
            future.get_loop().call_soon_threadsafe(future.set_result, to_return)

        elif isinstance(future, ConcurrentFuture):
            future.set_result(to_return)


on_request_node = csp.node(on_request)


def on_request_dict_basket(value: Dict[Any, ts[object]], future: ts[object]):  # type: ignore[no-untyped-def]
    with csp.start():
        csp.make_passive(value)

    if csp.ticked(future):
        # TODO consider returning something other than None?
        to_return = dict(value.validitems())  # type: ignore[attr-defined]

        if isinstance(future, AsyncFuture):
            future.get_loop().call_soon_threadsafe(future.set_result, to_return)

        elif isinstance(future, ConcurrentFuture):
            future.set_result(to_return)


on_request_node_dict_basket = csp.node(on_request_dict_basket)


def wait_for_next(value: ts[object], future: ts[object]):
    with csp.state():
        s_futures = []

    with csp.stop():
        # NOTE: resolve all dangling futures on shutdown
        for future in s_futures:
            if isinstance(future, AsyncFuture):
                # TODO use last value?
                future.get_loop().call_soon_threadsafe(future.set_result, None)

            elif isinstance(future, ConcurrentFuture):
                future.set_result(None)

    if csp.ticked(future):
        # save for later
        s_futures.append(future)

    if csp.ticked(value):
        for future in s_futures:
            if isinstance(future, AsyncFuture):
                future.get_loop().call_soon_threadsafe(future.set_result, value)

            elif isinstance(future, ConcurrentFuture):
                future.set_result(value)
        s_futures = []


wait_for_next_node = csp.node(wait_for_next)


def wait_for_next_dict_basket(value: Dict[Any, ts[object]], future: ts[object]):
    with csp.state():
        s_futures = []

    with csp.stop():
        # NOTE: resolve all dangling futures on shutdown
        for future in s_futures:
            if isinstance(future, AsyncFuture):
                # TODO use last value?
                future.get_loop().call_soon_threadsafe(future.set_result, None)

            elif isinstance(future, ConcurrentFuture):
                future.set_result(None)

    if csp.ticked(future):
        # save for later
        s_futures.append(future)

    if csp.ticked(value):
        # TODO consider returning something other than None?
        to_return = dict(value.validitems())  # type: ignore[attr-defined]

        for future in s_futures:
            if isinstance(future, AsyncFuture):
                future.get_loop().call_soon_threadsafe(future.set_result, to_return)

            elif isinstance(future, ConcurrentFuture):
                future.set_result(to_return)
        s_futures = []


wait_for_next_node_dict_basket = csp.node(wait_for_next_dict_basket)


def named_on_request_node(name: str) -> Any:
    # TODO this should work
    # csp.node(name=name)(on_request)
    node = csp.node(on_request)
    node.__name__ = name
    return node


def named_wait_for_next_node(name: str) -> Any:
    # TODO this should work
    # csp.node(name=name)(wait_for_next)
    node = csp.node(wait_for_next)
    node.__name__ = name
    return node


def named_on_request_node_dict_basket(name: str) -> Any:
    node = csp.node(on_request_dict_basket)
    node.__name__ = name
    return node


def named_wait_for_next_node_dict_basket(name: str) -> Any:
    node = csp.node(wait_for_next_dict_basket)
    node.__name__ = name
    return node


class AsyncFutureAdapter(FutureAdapter):
    def __init__(self, loop: AbstractEventLoop = None, name: str = None) -> None:
        super().__init__(kind="async", loop=loop, name=name)


class ConcurrentFutureAdapter(FutureAdapter):
    def __init__(self, name: str = None) -> None:
        super().__init__(kind="concurrent", name=name)



from datetime import datetime
from typing import Generic, List, Optional, Type

from ccflow import BaseModel
from csp.impl.enum import Enum
from csp.impl.genericpushadapter import GenericPushAdapter
from pydantic import Field

from csp_gateway.utils import get_dict_basket_key_type

from .channels import ChannelsType
from .module import Module


class ChannelsFactory(BaseModel, Generic[ChannelsType]):
    model_config = dict(arbitrary_types_allowed=True)  # (for FeedbackOutputDef)

    modules: List[Module[ChannelsType]] = Field(
        default_factory=list, description="The list of modules that will operate on the channels to build the csp graph."
    )
    channels: ChannelsType = Field(
        default=None,
        description="An instance of the Channels to be provided by the user for the Gateway. One can think of this as the internal message bus topics for the Gateway, "
        "even though there is no message bus, just named edges in a csp graph.",
    )
    channels_model: Type[ChannelsType] = Field(  # type: ignore[misc]
        default=ChannelsType,
        description="The type of the channels. Users of a `Gateway` are expected to pass `channels`, and `channels_model` will"
        "be automatically inferred from the type. Developers can subclass `Gateway` and set the default value of"
        "`channels_model` to be the specific type of channels that users must provide.",
    )
    block_set_channels_until: Optional[datetime] = Field(
        default=None,
        description="""
        This determines the csp time at which modules can start sending data to channels.
        This can be overriden on a per module basis, to allow some modules to send data to channels.
        """,
    )

    def build(self, channels: ChannelsType) -> ChannelsType:
        # First update the channels model type so we use the correct type,
        # important because we'll actually want to subclass the base GatewayChannels
        self.channels_model = channels.__class__

        # Collect dynamic keys from the channels
        if dynamic_keys := channels.dynamic_keys():
            for field, keys in dynamic_keys.items():
                channels._dynamic_keys[field].update((k, None) for k in keys)

        enabled_modules = [node for node in self.modules if not node.disable]
        # Collect dynamic keys for each node
        for node in enabled_modules:
            if dynamic_keys := node.dynamic_keys():
                for field, keys in dynamic_keys.items():
                    channels._dynamic_keys[field].update((k, None) for k in keys)

        if self.block_set_channels_until is not None:
            channels._block_set_channels_until = self.block_set_channels_until

        # Wire in each edge Provider.
        # The implementation of set_channel will handle multiplexing streams
        for node in enabled_modules:
            with channels._connection_context(node):
                node.connect(channels)

        # Now wire in the signals
        # first pass is for any baskets
        for (field, _indexer), push_adapter in channels._send_channels.items():
            if isinstance(push_adapter, tuple):
                # dict basket, plug in now that we should know all the
                # possible keys
                key_type = get_dict_basket_key_type(channels.get_outer_type(field))
                if isinstance(key_type, type) and issubclass(key_type, Enum):
                    for enumfield in key_type:
                        channels.add_send_channel(field, enumfield)

                    # add for whole basket ticks
                    channels._add_send_channel_dict_basket(field, key_type)
                else:
                    for key in channels._dynamic_keys.get(field, []):
                        channels.add_send_channel(field, key)

                    # add for whole basket ticks
                    channels._add_send_channel_dict_basket(field, channels._dynamic_keys.get(field, []))

        # second pass to finish connecting in wires
        for (field, indexer), push_adapter in channels._send_channels.items():
            with channels._connection_context(f"Send[{field}]{f'<{indexer}>' if indexer else ''}"):
                # Add it as an edge on the StreamGroup
                if isinstance(push_adapter, GenericPushAdapter):
                    channels.set_channel(field, push_adapter.out(), indexer=indexer)
                else:
                    # basket, first item of tuple is generic channel,
                    # second item is output
                    channels.set_channel(field, push_adapter[1], indexer=indexer)

        # Do any post work thats necessary
        channels._finalize()
        return channels



from json import dumps
from typing import Dict, List

from fastapi import Request
from fastapi.responses import HTMLResponse

from csp_gateway.server import GatewayChannels, GatewayModule

# separate to avoid circular
from csp_gateway.server.web import GatewayWebApp


class MountChannelsGraph(GatewayModule):
    route: str = "/channels_graph"

    def connect(self, channels: GatewayChannels) -> None:
        # NO-OP
        ...

    def rest(self, app: GatewayWebApp) -> None:
        api_router = app.get_router("api")
        app_router = app.get_router("app")

        # TODO subselect
        @api_router.get(
            self.route,
            response_model=Dict[str, Dict[str, List[str]]],
            tags=["Utility"],
        )
        def channels_graph_data(request: Request) -> Dict[str, Dict[str, List[str]]]:
            """
            This endpoint returns the structure of the GatewayChannels graph as a JSON object.
            It is used by the `Browse Channels Graph` endpoint to generate a nice, interactive view of the graph.

            Data is of the form:

            ```
            {
                "<channel name>": {
                    "getters": [`GatewayModule`s that pull from that channel],
                    "setters": [`GatewayModule`s that push to that channel]
                },
                ...
            }
            ```
            """
            return request.app.gateway.channels.graph()

        @app_router.get("/channels_graph", response_class=HTMLResponse, tags=["Utility"])
        def browse_channels_graph(request: Request):
            """
            This endpoint is a small webpage that shows the dependency relationship of the GatewayChannels graph powering this API.
            """
            channels_graph = request.app.gateway.channels.graph()
            return app.templates.TemplateResponse(
                "channels_graph.html.j2",
                {"request": request, "channels_graph": dumps(channels_graph)},
            )



import abc
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple, Union

import csp
from csp import ts
from csp.adapters.status import Status
from pydantic import Field

from csp_gateway.server.gateway.csp import ChannelsType
from csp_gateway.server.gateway.gateway import GatewayModule
from csp_gateway.utils import ReadWriteMode, set_alarm_and_fetch_alarm_time

from . import ChannelSelection, EncodedEngineCycle, JSONConverter

__all__ = ("EngineReplay",)


class EngineReplay(GatewayModule, abc.ABC):
    """
    This module defines an abstract class that allows reading and writing
    csp engine cycle ticks of the channels within a gateway.

    **WRITE**
    When writing, every tick of the specified channels are encoded and written to some source.

    **READ**
    From a specified source, the encodings for the ticks of the specified channels are decoded, then propagated through the underlying csp graph of the gateway.
    """

    selection: ChannelSelection = Field(default_factory=ChannelSelection)
    encode_selection: Optional[ChannelSelection] = Field(
        default=None,
        description=("Optional selection that can be specified to override the selection to specify channels only for encoding."),
    )
    decode_selection: Optional[ChannelSelection] = Field(
        default=None,
        description=("Optional selection that can be specified to override the selection to specify channels only for decoding."),
    )
    read_write_mode: ReadWriteMode = ReadWriteMode.READ_AND_WRITE
    requires: Optional[ChannelSelection] = []
    start_writing: Union[datetime, timedelta] = Field(
        default=timedelta(),
        description="""
        This determines the csp time at which the module stops reading. If in WRITE mode or
        READ_AND_WRITE mode, then the module will start writing the
        cycle encodings to Kafka at that time.
        If none is supplied, the module will start writing from the start of the graph.
        """,
    )
    flag_updates: Dict[str, List[Tuple[str, bool]]] = Field(
        default_factory=dict,
        description="""
        A mapping of channels to a list of tuples each containing:
        A field name and the flag it should be updated to, when in Read mode.
        Only works for non-dict basket channels
        """,
    )
    log_lagging_engine_cycles: bool = Field(
        default=False,
        description="Whether we log when engine cycles are lagging from the current engine time.",
    )
    subscribe_with_struct_id: bool = Field(
        default=True,
        description=("If False, replaces the id field on GatewayStructs from the source with one autogenerated by the current Gateway."),
    )
    subscribe_with_struct_timestamp: bool = Field(
        default=True,
        description=("If False, replaces the timestamp field on the GatewayStruct with a timestamp autogenerated by the current Gateway."),
    )

    @abc.abstractclassmethod
    def subscribe(self) -> Union[ts[str], ts[Dict[str, Any]]]:
        """Subscribe to engine ticks from some source.
        Returns ts that will be decoded into an engine cycle and replayed in the current gateway."""
        ...

    @abc.abstractclassmethod
    def publish(self, encoded_channels: ts[EncodedEngineCycle]) -> None:
        """Publish to some source an ts of an EncodedEngineCycle representing an encoding of a csp engine cycle."""
        ...

    def connect(self, channels: ChannelsType):
        """:meta private:"""
        decode_channels_list = encode_channels_list = self.selection.select_from(channels)
        if self.decode_selection is not None:
            decode_channels_list = self.decode_selection.select_from(channels)

        if self.encode_selection is not None:
            encode_channels_list = self.encode_selection.select_from(channels)

        exclude = set()
        if not self.subscribe_with_struct_id:
            exclude.add("id")
        if not self.subscribe_with_struct_timestamp:
            exclude.add("timestamp")

        json_channel_converter = JSONConverter(
            channels=channels,
            decode_exclude_fields=exclude,
            decode_channels=decode_channels_list,
            encode_channels=encode_channels_list,
            flag_updates=self.flag_updates,
            log_lagging_engine_cycles=self.log_lagging_engine_cycles,
        )

        writing_time_info = set_alarm_and_fetch_alarm_time(self.start_writing)

        # DECODING
        if self.read_write_mode in [ReadWriteMode.READ, ReadWriteMode.READ_AND_WRITE]:
            fat_pipe = self.subscribe()
            read = csp.times(fat_pipe) < writing_time_info.alarm_time
            fat_pipe_filtered = csp.filter(read, fat_pipe)
            json_channel_converter.decode(fat_pipe_filtered)

        # ENCODING
        if self.read_write_mode in [ReadWriteMode.WRITE, ReadWriteMode.READ_AND_WRITE]:
            encoded_channels = json_channel_converter.encode(writing_time_info.alarm_ticked)
            self.publish(encoded_channels=encoded_channels)

    def status(self) -> ts[Status]:
        "Status must be expicitly defined. Should return a ts of csp.Status structs that represent the status of this module."
        raise NotImplementedError()



import json
from typing import List, Tuple, TypeVar

import csp
from csp import ts
from csp_adapter_symphony import SymphonyAdapter, SymphonyMessage

from csp_gateway.server import GatewayChannels, GatewayModule

T = TypeVar("T")


class PublishSymphony(GatewayModule):
    """
    Takes a set of channels (selections) and turns them into symphony messages.

    To get setup with symphony:
    1. Email 'Tech - Symphony Admin' to create a new symphony account for your bot and for them to generate a pfx keyc.
        You should include the following in your email:
            * Bot Display Name (AAAA Alerts)
            * Distribution Email Address (e.g. trading-AAAA@cubistsystematic.com)
    2. Create a key.pem and crt.pem file locally (these will be used to authenticate your bot on your machine of choice):
    ( replace pfx_file with your filename)

    ```bash
        PFX="mypfxfile.pfx"
        openssl pkcs12 -in $PFX -out key.pem -nocerts -nodes
        openssl pkcs12 -in $PFX -out crt.pem -clcerts -nokeys
    ```

    3. On your target machine move your newly created key.pem and crt.pem files to an accessible directory.

    Args
    -----
    cert_path: str
        path to the crt.pem file (absolute)
    key_path: str
        path to the key.pem file (absolute)
    room_name: str
        name of the chat room with your bot
    user: str
        display name of your bot (from above)
    selections: List[str]
        list of channels to push to your symphony room

    Configuration
    -------------
    To be included in your watchtower config.

    ```yaml
    symphony_alerts:
        _target_: csp_gateway.PublishSymphony
        cert_path: /my/path/to/crt.pem
        key_path: /my/path/to/key.pem
        room_name: "My Symphony Room"
        user: "My Symphony Bot"
    ```

    """

    cert_path: str
    key_path: str
    room_name: str
    user: str

    selections: List[str] = []

    def get_cert_and_key(self) -> Tuple[str, str]:
        with open(self.cert_path, "r") as f:
            cert = f.read()
        with open(self.key_path, "r") as f:
            key = f.read()
        return cert, key

    def connect(self, channels: GatewayChannels):
        cert, key = self.get_cert_and_key()
        symphony_manager = SymphonyAdapter(cert, key)
        for channel in self.selections:
            sub = self.subscribe_channel(channels.get_channel(channel))
            symphony_manager.publish(csp.unroll(sub.messages))

    @csp.node
    def subscribe_channel(self, channel: ts["T"]) -> csp.Outputs(messages=ts[List[SymphonyMessage]]):
        if csp.ticked(channel):
            msgs = []
            # TODO: Not sure how to manage a dict basket here
            if isinstance(channel, list):
                msgs.extend(
                    [
                        SymphonyMessage(
                            user=self.user,
                            room=self.room_name,
                            msg=json.dumps(c.to_dict(), default=str),
                        )
                        for c in channel
                    ]
                )
            else:
                msgs.append(
                    SymphonyMessage(
                        user=self.user,
                        room=self.room_name,
                        msg=json.dumps(channel.to_dict(), default=str),
                    )
                )
            csp.output(messages=msgs)



from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, Dict, Union

import csp
from csp import ts

from csp_gateway.server import EncodedEngineCycle
from csp_gateway.server.shared.engine_replay import EngineReplay

from .json_pull_adapter import JSONPullAdapter

__all__ = ("ReplayEngineJSON",)


class JSONWriterThread(Thread):
    def __init__(self, file_path: str, write_queue: Queue, file_mode: str):
        super().__init__()
        self.file_path = file_path
        self.write_queue = write_queue
        self.file_mode = file_mode
        self.daemon = True

    def run(self):
        p = Path(self.file_path)
        p.parent.mkdir(parents=True, exist_ok=True)
        with p.open(self.file_mode) as f:
            while True:
                data = self.write_queue.get()
                f.write(data)
                f.flush()
                self.write_queue.task_done()


class ReplayEngineJSON(EngineReplay):
    overwrite_if_writing: bool = False
    filename: str

    @csp.node
    def _dump_to_json(self, to_store: ts[EncodedEngineCycle]):
        with csp.state():
            s_json_writer_thread = None
            s_queue = None

        with csp.start():
            file_mode = "w" if self.overwrite_if_writing else "a"
            s_queue = Queue()
            s_json_writer_thread = JSONWriterThread(self.filename, s_queue, file_mode)
            s_json_writer_thread.start()

        with csp.stop():
            s_queue.join()

        if csp.ticked(to_store):
            s_queue.put(to_store.encoding)

    def subscribe(self) -> Union[ts[str], ts[Dict[str, Any]]]:
        return JSONPullAdapter(self.filename)

    def publish(self, encoded_channels: ts[EncodedEngineCycle]) -> None:
        self._dump_to_json(encoded_channels)



import os
from datetime import datetime, timedelta
from typing import List

import csp
import sqlalchemy as db
from csp import ts
from sqlalchemy import MetaData, create_engine

from csp_gateway import Gateway, GatewayChannels, GatewayStruct
from csp_gateway.server.modules.sql import (
    ChannelSchemaConfig,
    PublishSQLA,
    SQLACnxDetails,
)
from csp_gateway.testing.harness import GatewayTestHarness


class MyTestStruct(GatewayStruct):
    a: str
    b: int


class MyTestStruct2(GatewayStruct):
    t: MyTestStruct


class GWC(GatewayChannels):
    test_channel: ts[MyTestStruct] = None
    test_channel_2d: ts[List[MyTestStruct]] = None
    test_channel_struct2: ts[MyTestStruct2] = None


def test_sqla_writer():
    open("db.sqlite", "w")  # create a fresh db to mess with
    cnx_details = SQLACnxDetails(engine="sqlite", host="db.sqlite")
    engine = create_engine(cnx_details.get_cnx_string())
    metadata = MetaData()
    test_table = db.Table(
        "test_struct_table",
        metadata,
        db.Column("not_a", db.String(255), nullable=True),
        db.Column("b", db.Integer, nullable=True),
        db.Column("c", db.String(255), nullable=True),
    )
    schema_configs = [
        ChannelSchemaConfig(
            channel_name="test_channel",
            table="test_struct_table",
            fields=["a", "b"],
            rename_fields={"a": "not_a"},
            augmentation_fields={"c": "I AM TESTOR"},
        )
    ]
    sqla_writer = PublishSQLA(cnx_details=cnx_details, schema_configs=schema_configs)
    metadata.create_all(engine)

    h = GatewayTestHarness(test_channels=[GWC.test_channel])

    t = MyTestStruct(timestamp=datetime(2023, 1, 1), a="a", b=1)
    h.send(GWC.test_channel, t)
    h.assert_ticked(GWC.test_channel, 1)

    gateway = Gateway(modules=[h, sqla_writer], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(1))

    # kind of tricky but this works
    with engine.connect() as connection:
        res = connection.execute(test_table.select()).fetchone()
        assert res
    assert res.not_a == t.a
    assert res.b == t.b
    assert res.c == "I AM TESTOR"

    os.remove("db.sqlite")


def test_struct_of_struct():
    open("db.sqlite", "w")  # create a fresh db to mess with
    cnx_details = SQLACnxDetails(engine="sqlite", host="db.sqlite")
    engine = create_engine(cnx_details.get_cnx_string())
    metadata = MetaData()
    test_table = db.Table(
        "test_struct_table",
        metadata,
        db.Column("t", db.String(255), nullable=True),
    )
    schema_configs = [
        ChannelSchemaConfig(
            channel_name="test_channel_struct2",
            table="test_struct_table",
            fields=["t"],
        )
    ]
    sqla_writer = PublishSQLA(cnx_details=cnx_details, schema_configs=schema_configs)
    metadata.create_all(engine)

    h = GatewayTestHarness(test_channels=[GWC.test_channel_struct2])

    t = MyTestStruct(timestamp=datetime(2023, 1, 1), a="a", b=1)
    t2 = MyTestStruct2(t=t)
    h.send(GWC.test_channel_struct2, t2)
    h.assert_ticked(GWC.test_channel_struct2, 1)

    gateway = Gateway(modules=[h, sqla_writer], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(1))

    # kind of tricky but this works
    with engine.connect() as connection:
        res = connection.execute(test_table.select()).fetchone()
    assert res is not None

    os.remove("db.sqlite")


def test_sqla_writer_2d():
    open("db.sqlite", "w")  # create a fresh db to mess with
    cnx_details = SQLACnxDetails(engine="sqlite", host="db.sqlite")
    engine = create_engine(cnx_details.get_cnx_string())
    metadata = MetaData()
    test_table = db.Table(
        "test_struct_table",
        metadata,
        db.Column("not_a", db.String(255), nullable=True),
        db.Column("b", db.Integer, nullable=True),
        db.Column("c", db.String(255), nullable=True),
    )
    schema_configs = [
        ChannelSchemaConfig(
            channel_name="test_channel_2d",
            table="test_struct_table",
            fields=["a", "b"],
            rename_fields={"a": "not_a"},
            augmentation_fields={"c": "I AM TESTOR"},
        )
    ]
    sqla_writer = PublishSQLA(cnx_details=cnx_details, schema_configs=schema_configs)
    metadata.create_all(engine)

    h = GatewayTestHarness(test_channels=[GWC.test_channel_2d])

    t1 = MyTestStruct(timestamp=datetime(2023, 1, 1), a="a", b=1)
    t2 = MyTestStruct(timestamp=datetime(2023, 1, 1), a="a", b=2)
    h.send(GWC.test_channel_2d, [t1, t2])
    h.assert_ticked(GWC.test_channel_2d, 1)

    gateway = Gateway(modules=[h, sqla_writer], channels=GWC())
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(1))

    # kind of tricky but this works
    with engine.connect() as connection:
        res = connection.execute(test_table.select()).fetchall()
    assert res[0].not_a == t1.a
    assert res[0].b == t1.b
    assert res[0].c == "I AM TESTOR"
    assert res[1].not_a == t2.a
    assert res[1].b == t2.b
    assert res[1].c == "I AM TESTOR"

    os.remove("db.sqlite")



import json
from datetime import datetime

from csp_gateway.server.demo import ExampleData
from csp_gateway.utils.web.filter import Filter, FilterCondition, FilterWhere, FilterWhereLambdaMap
from csp_gateway.utils.web.query import Query

DUMMY_STATE_DATA = [
    {
        "id": "2319519293118611459",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:20.395000+00:00"),
        "x": 1,
        "y": "111",
    },
    {
        "id": "2319519293118611463",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:21.396000+00:00"),
        "x": 2,
        "y": "222",
    },
    {
        "id": "2319519293118611467",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:22.394000+00:00"),
        "x": 3,
        "y": "333",
    },
    {
        "id": "2319519293118611471",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:23.394000+00:00"),
        "x": 4,
        "y": "444",
    },
    {
        "id": "2319519293118611475",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:24.394000+00:00"),
        "x": 5,
        "y": "555",
    },
    {
        "id": "2319519293118611479",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:25.394000+00:00"),
        "x": 6,
        "y": "666",
    },
    {
        "id": "2319519293118611483",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:26.394000+00:00"),
        "x": 7,
        "y": "777",
    },
    {
        "id": "2319519293118611487",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:27.394000+00:00"),
        "x": 8,
        "y": "888",
    },
    {
        "id": "2319519293118611491",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:28.394000+00:00"),
        "x": 9,
        "y": "999",
    },
    {
        "id": "2319519293118611495",
        "timestamp": datetime.fromisoformat("2023-03-30T14:45:29.394000+00:00"),
        "x": 0,
        "y": "101010",
    },
]

DUMMY_STATE_DATA = [ExampleData(**d) for d in DUMMY_STATE_DATA]


class TestQuery:
    def test_serialize_deserialize(self):
        filters = [
            Filter(attr="f1", by=FilterCondition(value="123", where="==")),
            Filter(attr="timestamp", by=FilterCondition(when=datetime(2000, 1, 1), where="==")),
        ]
        for f in filters:
            assert f.model_dump_json() == Filter.model_validate_json(f.model_dump_json()).model_dump_json()

    def test_lambda_map(self):
        for key in FilterWhere.__args__:
            assert key in FilterWhereLambdaMap

    def test_query_null(self):
        q = Query()
        assert q.calculate(DUMMY_STATE_DATA) == DUMMY_STATE_DATA
        q = Query(filters=[])
        assert q.calculate(DUMMY_STATE_DATA) == DUMMY_STATE_DATA

    def test_query_bad(self):
        q = Query(filters=[Filter(attr="", by=FilterCondition(value="", where="=="))])
        assert q.calculate(DUMMY_STATE_DATA) == []

        # TODO more

    def test_query_simple(self):
        q = Query(filters=[Filter(attr="x", by=FilterCondition(value=1, where="=="))])
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.x == 1]

        q = Query(filters=[Filter(attr="x", by=FilterCondition(value=5, where="<="))])
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.x <= 5]

        q = Query(
            filters=[
                Filter(attr="x", by=FilterCondition(value=5, where="<=")),
                Filter(attr="x", by=FilterCondition(value=2, where=">=")),
            ]
        )
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.x <= 5 and d.x >= 2]

        q = Query(filters=[Filter(attr="x", by=FilterCondition(value=0, where="=="))])
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.x == 0]

    def test_query_attr(self):
        q = Query(filters=[Filter(attr="id", by=FilterCondition(attr="y", where=">="))])
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.id >= d.y]

    def test_query_when(self):
        when = datetime.fromisoformat("2023-03-30T14:45:24.394000+00:00")
        q = Query(filters=[Filter(attr="timestamp", by=FilterCondition(when=when, where="=="))])
        assert q.calculate(DUMMY_STATE_DATA) == [d for d in DUMMY_STATE_DATA if d.timestamp == when]


class TestFilterCondition:
    def test_query_loading(self):
        filter_str = json.dumps({"value": 123, "where": "=="})
        f = FilterCondition.model_validate_json(filter_str)
        assert isinstance(f.value, int)
        assert f.value == 123

        filter_str = json.dumps({"value": 123.0, "where": "=="})
        f = FilterCondition.model_validate_json(filter_str)
        assert isinstance(f.value, float)
        assert f.value == 123.0

        filter_str = json.dumps({"value": 123.5, "where": "=="})
        f = FilterCondition.model_validate_json(filter_str)
        assert isinstance(f.value, float)
        assert f.value == 123.5

        filter_str = json.dumps({"value": "123.5", "where": "=="})
        f = FilterCondition.model_validate_json(filter_str)
        assert isinstance(f.value, str)
        assert f.value == "123.5"



import getpass
import resource
import socket
import threading
from datetime import datetime, timezone

import csp
import psutil
from csp import ts

from csp_gateway.server import GatewayChannels, GatewayModule

# separate to avoid circular
from csp_gateway.server.web import GatewayWebApp
from csp_gateway.utils import Controls

_HOSTNAME = socket.gethostname()
_USER = getpass.getuser()


class MountControls(GatewayModule):
    mount_heartbeat: bool = True
    mount_stats: bool = True
    mount_shutdown: bool = True

    def connect(self, channels: GatewayChannels) -> None:
        self.subscribe(channels.get_channel("controls"))
        channels.add_send_channel("controls")

    def rest(self, app: GatewayWebApp) -> None:
        available = []
        if self.mount_heartbeat:
            app.add_controls_api(field="heartbeat")
            available.append("heartbeat")
        if self.mount_stats:
            app.add_controls_api(field="stats")
            available.append("stats")
        if self.mount_shutdown:
            app.add_controls_api(field="shutdown")
            available.append("shutdown")

        app.add_controls_available_channels(fields=set(available))

    @csp.node
    def manage_controls(self, data: ts[Controls]):
        if csp.ticked(data):
            # TODO better check if "seen"
            if data.name == "heartbeat":
                # don't have to do anything
                data.status = "ok"

            elif data.name == "stats" and not data.data:
                stats = {}

                # Machine information
                stats["cpu"] = psutil.cpu_percent()
                stats["memory"] = psutil.virtual_memory().percent
                stats["memory-total"] = round(
                    psutil.virtual_memory().available * 100 / psutil.virtual_memory().total,
                    2,
                )

                # Process and thread information
                current_process = psutil.Process()
                stats["pid"] = current_process.pid
                stats["active_threads"] = threading.active_count()

                # Get max threads from ulimit
                _, hard_limit = resource.getrlimit(resource.RLIMIT_NPROC)
                stats["max_threads"] = hard_limit if hard_limit != resource.RLIM_INFINITY else "unlimited"

                # Time information
                stats["now"] = datetime.now(timezone.utc)
                stats["csp-now"] = csp.now()

                stats["host"] = _HOSTNAME
                stats["user"] = _USER

                data.data = stats
                data.status = "ok"  # we mark as ok at the end only after we have all the data

    @csp.graph
    def subscribe(self, data: ts[Controls]):
        self.manage_controls(data)



from .channels import Channels, ChannelsType
from .factory import ChannelsFactory
from .futures import (
    AsyncFutureAdapter,
    ConcurrentFutureAdapter,
    FutureAdapter,
    named_on_request_node,
    named_on_request_node_dict_basket,
    on_request_node,
    on_request_node_dict_basket,
)
from .module import Module
from .state import *



import logging
import timeit
from datetime import datetime, timezone
from unittest.mock import MagicMock, patch

import pytest

from csp_gateway.server.modules.logging.util import (
    DatadogLevel,
    MonitoringBase,
    MonitoringEvent,
    MonitoringLevelMapping,
    MonitoringMetric,
    MonitoringSystem,
    OpsGenieLevel,
)
from csp_gateway.utils.struct.base import GatewayStruct


# Mocking GatewayStruct for simplicity
class MockGatewayStruct(GatewayStruct):
    def __init__(self, **kwargs):
        self.id = kwargs.get("id", "test_id")
        self.timestamp = kwargs.get("timestamp", datetime.now(timezone.utc))


@pytest.fixture
def monitoring_event():
    return MonitoringEvent(
        id="event_id",
        timestamp=datetime.now(timezone.utc),
        title="Test Event",
        text="This is a test event",
        event_type="test_event",
        alert_type=DatadogLevel.error,
        source="test_source",
        tags={
            "env": "test",
            "event_group": "group1",
            "tag1": "value1",
            "tag2": "value2",
        },
    )


@pytest.fixture
def monitoring_metric():
    return MonitoringMetric(
        id="metric_id",
        timestamp=datetime.now(timezone.utc),
        metric="test_metric",
        metric_type=MonitoringMetric.MetricType.gauge,
        value=1.0,
        tags={"env": "test"},
    )


def test_monitoring_system_enum():
    assert MonitoringSystem.DATADOG.value == "datadog"
    assert MonitoringSystem.OPSGENIE.value == "opsgenie"
    assert MonitoringSystem.LOGGING.value == "logging"


def test_datadog_level_enum():
    assert DatadogLevel.error.value == "error"
    assert DatadogLevel.warning.value == "warning"
    assert DatadogLevel.info.value == "info"
    assert DatadogLevel.success.value == "success"


def test_opsgenie_level_enum():
    assert OpsGenieLevel.P1.value == "P1"
    assert OpsGenieLevel.P2.value == "P2"
    assert OpsGenieLevel.P3.value == "P3"
    assert OpsGenieLevel.P4.value == "P4"
    assert OpsGenieLevel.P5.value == "P5"


def test_monitoring_base_init():
    base = MonitoringBase(id="base_id", timestamp=datetime.now(timezone.utc), tags={"env": "test"})
    assert base.id == "base_id"
    assert base.tags == {"env": "test"}


def test_monitoring_base_get_tags(monitoring_event):
    tags = monitoring_event.get_tags()
    assert "env:test" in tags
    assert "monitoring_id:event_id" in tags


def test_monitoring_base_get_datadog_timestamp(monitoring_event):
    timestamp = monitoring_event.get_datadog_timestamp()
    assert isinstance(timestamp, float)


def test_monitoring_event_to_datadog(monitoring_event):
    payload = monitoring_event.to_datadog()
    assert payload["title"] == "Test Event"
    assert payload["text"] == "This is a test event"
    assert payload["alert_type"] == "error"


def test_monitoring_metric_to_datadog(monitoring_metric):
    payload = monitoring_metric.to_datadog()
    assert payload["metric"] == "test_metric"
    assert payload["type"] == "gauge"
    assert isinstance(payload["points"], list)


@patch("opsgenie_sdk.CreateAlertPayload")
def test_monitoring_event_to_opsgenie(mock_create_alert_payload, monitoring_event):
    # Set up the mock return value
    mock_payload_instance = MagicMock()
    mock_payload_instance.source = "test_source"
    mock_payload_instance.message = "Test Event"
    mock_create_alert_payload.return_value = mock_payload_instance

    payload = monitoring_event.to_opsgenie()
    assert mock_create_alert_payload.called
    assert payload.source == "test_source"
    assert payload.message == "Test Event"


@patch("opsgenie_sdk.CreateAlertPayload")
def test_monitoring_metric_to_opsgenie(mock_create_alert_payload, monitoring_metric):
    # Set up the mock return value
    mock_payload_instance = MagicMock()
    mock_payload_instance.source = "python"
    mock_payload_instance.message = "Test Metric Alert"
    mock_create_alert_payload.return_value = mock_payload_instance

    payload = monitoring_metric.to_opsgenie(
        title="Test Metric Alert",
        alert_type=MonitoringLevelMapping.info,
    )
    assert mock_create_alert_payload.called
    assert payload.source == "python"
    assert payload.message == "Test Metric Alert"


def test_monitoring_event_to_dict(monitoring_event):
    event_dict = monitoring_event.to_dict()
    assert event_dict["title"] == "Test Event"
    assert event_dict["text"] == "This is a test event"
    assert event_dict["event_type"] == "test_event"
    assert event_dict["alert_type"] == DatadogLevel.error
    assert event_dict["source"] == "test_source"
    assert event_dict["tags"] == [
        "env:test",
        "event_group:group1",
        "tag1:value1",
        "tag2:value2",
        "monitoring_id:event_id",
    ]
    assert "timestamp" in event_dict


def test_monitoring_metric_to_dict(monitoring_metric):
    metric_dict = monitoring_metric.to_dict()
    assert metric_dict["metric"] == "test_metric"
    assert metric_dict["type"] == "gauge"
    assert metric_dict["points"][0][1] == 1.0
    assert metric_dict["tags"] == ["env:test", "monitoring_id:metric_id"]


def test_monitoring_level_mapping_from_event(monitoring_event):
    mapping = MonitoringLevelMapping.from_event(monitoring_event)
    assert mapping == MonitoringLevelMapping.error
    assert mapping.value.datadog == DatadogLevel.error
    assert mapping.value.opsgenie == OpsGenieLevel.P2
    assert mapping.value.logging == logging.ERROR

    with pytest.raises(ValueError):

        class Foo:
            def __init__(self):
                self.alert_type = "foo"

        MonitoringLevelMapping.from_event(Foo())


def test_monitoring_levels():
    assert MonitoringLevelMapping.get_opsgenie_levels() == (
        "P1",
        "P2",
        "P3",
        "P4",
        "P5",
    )
    assert MonitoringLevelMapping.get_datadog_levels() == (
        "error",
        "error",
        "warning",
        "info",
        "success",
    )
    assert MonitoringLevelMapping.get_logging_levels() == (
        50,
        40,
        30,
        20,
        10,
    )


def test_monitoring_level_mapping_from_alert_type():
    # Datadog has no "critical" level alert
    for level in (OpsGenieLevel.P1, "P1", 50):
        mapping = MonitoringLevelMapping.from_alert_type(level)
        assert mapping == MonitoringLevelMapping.critical
        assert mapping.value.level == "critical"
        assert mapping.value.datadog == DatadogLevel.error
        assert mapping.value.logging == 50

    for level in (OpsGenieLevel.P2, DatadogLevel.error, "error", "P2", 40):
        mapping = MonitoringLevelMapping.from_alert_type(level)
        assert mapping == MonitoringLevelMapping.error
        assert mapping.value.level == "error"
        assert mapping.value.datadog == DatadogLevel.error
        assert mapping.value.logging == 40

    for level in (OpsGenieLevel.P3, DatadogLevel.warning, "warning", "P3", 30):
        mapping = MonitoringLevelMapping.from_alert_type(level)
        assert mapping == MonitoringLevelMapping.warning
        assert mapping.value.level == "warning"
        assert mapping.value.datadog == DatadogLevel.warning
        assert mapping.value.logging == 30

    for level in (OpsGenieLevel.P4, DatadogLevel.info, "info", "P4", 20):
        mapping = MonitoringLevelMapping.from_alert_type(level)
        assert mapping == MonitoringLevelMapping.info
        assert mapping.value.level == "info"
        assert mapping.value.datadog == DatadogLevel.info
        assert mapping.value.logging == 20

    for level in (OpsGenieLevel.P5, DatadogLevel.success, "success", "P5", 10):
        mapping = MonitoringLevelMapping.from_alert_type(level)
        assert mapping == MonitoringLevelMapping.debug
        assert mapping.value.level == "debug"
        assert mapping.value.datadog == DatadogLevel.success
        assert mapping.value.logging == 10

    with pytest.raises(ValueError):
        MonitoringLevelMapping.from_alert_type("foo")

    with pytest.raises(ValueError):
        MonitoringLevelMapping.from_alert_type(42)


def test_monitoring_event_class_methods():
    event = MonitoringEvent(
        id="event_id",
        timestamp=datetime.now(timezone.utc),
        title="Test Event",
        text="This is a test event",
        event_type="test_event",
        alert_type=DatadogLevel.error,
        source="test_source",
        tags={"env": "test"},
    )
    event_dict = event.to_dict()
    assert event_dict["title"] == "Test Event"
    assert event_dict["text"] == "This is a test event"
    flatten_data = event.psp_flatten()[0]
    assert '{"env":"test","monitoring_id":"event_id"}' == flatten_data["tag_str"]


def test_monitoring_metric_class_methods():
    metric = MonitoringMetric(
        id="metric_id",
        timestamp=datetime.now(timezone.utc),
        metric="test_metric",
        metric_type=MonitoringMetric.MetricType.gauge,
        value=1.0,
        tags={"env": "test"},
    )
    metric_dict = metric.to_dict()
    assert metric_dict["metric"] == "test_metric"
    assert metric_dict["type"] == "gauge"
    assert metric_dict["points"][0][1] == 1.0
    flatten_data = metric.psp_flatten()[0]
    assert '{"env":"test","monitoring_id":"metric_id"}' == flatten_data["tag_str"]


@patch("opsgenie_sdk.CreateAlertPayload")
def test_monitoring_event_alias(mock_create_alert_payload, monitoring_event):
    # Set up the mock return value
    mock_payload_instance = MagicMock()
    mock_payload_instance.source = "test_source"
    mock_payload_instance.message = "Test Event"
    mock_create_alert_payload.return_value = mock_payload_instance

    # Test case where alias is generated
    alias_tags = {"group1": ["tag2", "tag1"]}
    monitoring_event.to_opsgenie(alias_tags=alias_tags)
    assert mock_create_alert_payload.called

    # Verify the alias is correctly generated
    expected_tags = {
        "env:test",
        "monitoring_id:event_id",
        "event_group:group1",
        "tag1:value1",
        "tag2:value2",
    }
    actual_call_args = mock_create_alert_payload.call_args[1]
    assert set(actual_call_args["tags"]) == expected_tags
    assert actual_call_args["alias"] == "value2:value1"

    # Reset the mock for the next test
    mock_create_alert_payload.reset_mock()

    # Test case where alias is empty, and id is used instead
    alias_tags = {"group2": ["tag1", "tag2"]}
    monitoring_event.to_opsgenie(alias_tags=alias_tags)
    assert mock_create_alert_payload.called

    # Verify the alias is set to the event id
    actual_call_args = mock_create_alert_payload.call_args[1]
    assert set(actual_call_args["tags"]) == expected_tags
    assert actual_call_args["alias"] == "event_id"


def test_alias(monitoring_event):
    # Test default values
    result = monitoring_event.get_event_alias()
    assert result == "", "Default values should return an empty string"

    # Test successful execution with valid inputs
    alias_tags = {"group1": ["tag1", "tag2"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags)
    assert result == "value1:value2", "Expected 'value1:value2'"

    # Test that the order of tags is preserved as defined in the config
    alias_tags = {"group1": ["tag2", "tag1"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags)
    assert result == "value2:value1", "Expected 'value2:value1'"

    # Test when alias_tags is None
    result = monitoring_event.get_event_alias(alias_tags=None)
    assert result == "", "alias_tags=None should return an empty string"

    # Test when category_tag does not match any tags
    alias_tags = {"group2": ["tag1", "tag2"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags)
    assert result == "", "Non-matching category_tag should return an empty string"

    # Test when none of the alias requested is found
    alias_tags = {"group1": ["non_existent_tag1", "non_existent_tag2"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags)
    assert result == "", "None of the requested alias tags found should return an empty string"

    # Test when there are no tags
    monitoring_event.tags = {}
    alias_tags = {"group1": ["tag1", "tag2"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags)
    assert result == "", "No tags should return an empty string"
    monitoring_event.tags = {
        "env": "test",
        "event_group": "group1",
        "tag1": "value1",
        "tag2": "value2",
    }  # Reset tags for further tests

    # Test when the category tag is not found
    alias_tags = {"group1": ["tag1", "tag2"]}
    result = monitoring_event.get_event_alias(category_tag="non_existent_category", alias_tags=alias_tags)
    assert result == "", "Non-existent category tag should return an empty string"

    # Test when extra_tags are included
    alias_tags = {"group1": ["tag1", "tag2", "extra_tag"]}
    extra_tags = {"extra_tag": "extra_value"}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags, extra_tags=extra_tags)
    assert result == "value1:value2:extra_value", "Expected 'value1:value2:extra_value'"

    # Test with different separator
    alias_tags = {"group1": ["tag1", "tag2"]}
    result = monitoring_event.get_event_alias(alias_tags=alias_tags, separator="-")
    assert result == "value1-value2", "Expected 'value1-value2'"

    # Test the speed of execution for the function
    alias_tags = {"group1": ["tag1", "tag2"]}
    execution_time = timeit.timeit(lambda: monitoring_event.get_event_alias(alias_tags=alias_tags), number=1000)
    print(f"Execution time for 1000 runs: {execution_time} seconds")



from .adapter import (
    AsyncFutureAdapter,
    ConcurrentFutureAdapter,
    FutureAdapter,
    named_on_request_node,
    named_on_request_node_dict_basket,
    named_wait_for_next_node,
    named_wait_for_next_node_dict_basket,
    on_request_node,
    on_request_node_dict_basket,
    wait_for_next_dict_basket,
    wait_for_next_node,
)



import _thread
import itertools
from datetime import date, datetime
from enum import Enum as PyEnum
from logging import getLogger
from typing import Any, Callable, Dict, GenericAlias, List, Optional, Set, Tuple, Union, _GenericAlias, get_args, get_origin

import orjson
from csp import Enum, Struct
from csp.impl.enum import EnumMeta
from csp.impl.types.container_type_normalizer import ContainerTypeNormalizer
from numpy import ndarray

__all__ = (
    "CustomJsonifier",
    "psp_flatten_dict",
    "psp_flatten_list",
    "psp_flatten",
    "ExcludedColumns",
    "psp_schema",
    "PerspectiveUtilityMixin",
)

log = getLogger(__name__)

CustomJsonifier = Callable[[Any], Tuple[Any, bool]]


# We expose these functions separate from the class definition
# so that they can be called recursively.
# However, the top level call should always come from a
# a PerspectiveUtilityMixin subclass or instance of such.
def psp_flatten_dict(obj: Dict[str, Any]) -> Any:
    """Flatten dicts of values into a top level dict"""

    # Base template to use for creating copies later
    base_dict = {}
    # List of deltas that will be applied to the base dict
    # to create final dicts
    list_items = []
    for obj_key, obj_val in obj.items():
        # Flatten every item
        res = psp_flatten(obj_val)
        if isinstance(res, list):
            if len(res) == 0:
                # Key needs to be deleted, use empty dict as delta
                list_items.append([{}])

            else:
                # Delta options for current key
                delta_list = []
                for res_val in res:
                    # NOTE: We should never receive a non-empty list as an item here

                    # Delta dict for the current item in list
                    delta_dict = {}
                    if isinstance(res_val, dict):
                        # Merge the sub-dict with parent key
                        for k, v in res_val.items():
                            delta_dict["{}.{}".format(obj_key, k)] = v
                        delta_list.append(delta_dict)
                    elif isinstance(res_val, list) and len(res) == 0:
                        # Key needs to be deleted, use empty dict as delta
                        delta_list.append({})
                    else:
                        # Create new delta dict with key, val
                        delta_list.append({obj_key: res_val})
                list_items.append(delta_list)
        else:
            # Non-list item found, use as is
            base_dict[obj_key] = res

    ret = []
    # Process all possible combinations of delta options for keys
    for combination in itertools.product(*list_items):
        # Make a copy from the base template
        new_dict = base_dict.copy()
        for elem in combination:
            # Merge the delta into copy
            new_dict.update(elem)
        ret.append(new_dict)

    return ret


def psp_flatten_list(obj: List[Any]) -> List[Any]:
    """Flatten list of complex types (sub-lists, dicts) to a top level list"""

    ret = []
    for val in obj:
        res = psp_flatten(val)
        if isinstance(res, list) and res:
            # Flatten sub-lists into a single list
            # NOTE: Special handling for empty list
            #  Empty list indicate key should be deleted
            #  so we preserve empty list during flattening
            ret.extend(res)
        else:
            ret.append(res)
    return ret


def psp_flatten(obj: Any) -> Any:
    """Flatten an object"""

    #  This should only return simple objects or lists of simple objects (not dicts)
    ret = obj
    if isinstance(obj, list):
        ret = psp_flatten_list(obj)
    elif isinstance(obj, dict):
        ret = psp_flatten_dict(obj)
    return ret


ExcludedColumns = Union[Set[str], Dict[str, Union[bool, "ExcludedColumns"]]]


def _is_excluded(field: str, excluded_columns: ExcludedColumns) -> Union[bool, ExcludedColumns]:
    if isinstance(excluded_columns, set):
        return field in excluded_columns

    return excluded_columns.get(field, False)


def _is_optional(t: type) -> bool:
    if not isinstance(t, (GenericAlias, _GenericAlias)):
        return False
    if get_origin(t) is not Union:
        return False
    args = list(get_args(t))
    if len(args) != 2 or type(None) not in args:
        return False
    return True


def _get_type_from_optional(t: type):
    args = list(get_args(t))
    args.remove(type(None))
    return args[0]


def psp_schema(cls, excluded_columns: Optional[ExcludedColumns] = None) -> Dict[str, type]:
    """Returns the perspective schema for a class.

    Args:
        excluded_columns: Columns to exclude from the schema.
    """

    # Pydantic doesn't support fields that start with underscore
    schema = {k: v for k, v in cls.metadata(typed=False).items() if not k.startswith("_")}
    schema_annotated = {k: v for k, v in cls.metadata(typed=True).items() if not k.startswith("_")}
    add = {}
    remove = []

    for field, value in schema.items():
        # Make sure its a type so `issubclass`
        # calls don't fail
        if _is_optional(schema_annotated[field]):
            value = _get_type_from_optional(schema_annotated[field])
            value = ContainerTypeNormalizer.normalized_type_to_actual_python_type(value)

        if not isinstance(value, type):
            # TODO other generics
            if isinstance(value, list):
                value = value[0]
                schema[field] = value
            else:
                remove.append(field)
                # TODO deal with dropped
                log.warning(f"Type is not actually a type: {field} {value}")
                continue

        is_excluded = excluded_columns and _is_excluded(field, excluded_columns)
        if is_excluded:
            remove.append(field)

        if issubclass(value, list) or issubclass(value, ndarray):
            try:
                # will be unrolled into root type
                try:
                    annotation = cls.__annotations__[field]
                except KeyError:
                    # Fallback: search bases for annotation
                    annotation = None
                    for base in cls.__mro__[1:]:
                        anns = getattr(base, "__annotations__", {})
                        if field in anns:
                            annotation = anns[field]
                            break
                    if annotation is None:
                        raise KeyError(field)

                # get arg type
                arg = get_args(annotation)[0]

                # use this as type
                value = arg
            except (KeyError, IndexError):
                # just use str
                value = str
            finally:
                schema[field] = value

        # If its a complicated type that we just serialize to json, leave as str
        if issubclass(value, dict):
            schema[field] = str
            continue

        # If its an enum, promote to str
        if issubclass(value, (Enum, EnumMeta, PyEnum)):
            schema[field] = str
            continue

        # Otherwise if its not a handled type
        if (
            not issubclass(value, str)
            and not issubclass(value, int)
            and not issubclass(value, float)
            and not issubclass(value, bool)
            and not issubclass(value, datetime)
            and not issubclass(value, date)
        ):
            excluded_sub_fields = None
            if is_excluded:
                # no need to add field to remove, it has been added already
                if isinstance(is_excluded, bool):
                    if is_excluded:
                        continue

                else:
                    excluded_sub_fields = is_excluded

            else:
                # remove it from the schema
                remove.append(field)

            # if its a struct, flatten
            if issubclass(value, Struct):
                if hasattr(value, "psp_schema"):
                    struct_items = value.psp_schema(excluded_sub_fields).items()
                else:
                    struct_items = psp_schema(value, excluded_sub_fields).items()

                # add subschema
                for subkey, subvalue in struct_items:
                    add["{}.{}".format(field, subkey)] = subvalue
            else:
                # TODO deal with dropped
                log.warning(f"Type {value} on has no perspective conversion, ignoring in perspective tables: {cls.__name__}.{field}")

    # remove all that need to be removed
    for to_remove in remove:
        schema.pop(to_remove)

    for key in schema.keys():
        if schema[key] is object and _is_optional(schema_annotated[key]):
            schema[key] = _get_type_from_optional(schema_annotated[key])

    schema.update(add)
    return schema


class PerspectiveUtilityMixin:
    def psp_flatten(self, custom_jsonifier: Optional[CustomJsonifier] = None) -> List[Dict[str, Any]]:
        def _callback(obj):
            """Callback helper that either calls custom_jsonifier or a default set of conversions"""
            if custom_jsonifier:
                obj = custom_jsonifier(obj)
            if isinstance(obj, ndarray):
                return obj.tolist()
            elif isinstance(obj, set):
                return list(obj)
            elif isinstance(obj, (PyEnum, Enum)):
                return obj.name
            elif isinstance(obj, _thread.LockType):
                return "<Lock>"
            else:
                log.warning(f"No serializer for {obj}, converting to ''")
                return ""

        json_obj = orjson.loads(self.to_json(_callback))
        flat_obj = psp_flatten(json_obj)
        return flat_obj

    @classmethod
    def psp_schema(cls, excluded_columns: Optional[ExcludedColumns] = None) -> Dict[str, type]:
        """Return the perspective schema.

        Args:
            excluded_columns: Columns to exclude from the schema.
        """
        return psp_schema(cls, excluded_columns)



import logging

from pydantic import Field, model_validator

from csp_gateway.server import ChannelSelection, GatewayChannels, GatewayModule

# separate to avoid circular
from csp_gateway.server.web import GatewayWebApp

log = logging.getLogger(__name__)


class MountRestRoutes(GatewayModule):
    force_mount_all: bool = Field(
        False,
        description="For debugging, will mount all rest routes for every channel, including state and send routes, even if not added by any modules",
    )

    mount_last: ChannelSelection = Field(default_factory=ChannelSelection, description="Channels to mount for last operations. Defaults to all")
    mount_next: ChannelSelection = Field(default_factory=ChannelSelection, description="Channels to mount for next operations. Defaults to all")
    mount_send: ChannelSelection = Field(default_factory=ChannelSelection, description="Channels to mount for send operations. Defaults to all")
    mount_state: ChannelSelection = Field(default_factory=ChannelSelection, description="Channels to mount for state operations. Defaults to all")
    mount_lookup: ChannelSelection = Field(default_factory=ChannelSelection, description="Channels to mount for lookup operations. Defaults to all")

    @model_validator(mode="before")
    @classmethod
    def _deprecate_mount_all(cls, values):
        if (mount_all := values.pop("mount_all", None)) is not None:
            log.warning("mount_all is deprecated, please use force_mount_all instead")
            if "force_mount_all" not in values:
                values["force_mount_all"] = mount_all
        return values

    def connect(self, channels: GatewayChannels) -> None:
        # NO-OP
        ...

    def rest(self, app: GatewayWebApp) -> None:
        self._mount_last(app)
        self._mount_next(app)
        self._mount_send(app)
        self._mount_state(app)
        self._mount_lookup(app)

    def _mount_last(self, app: GatewayWebApp) -> None:
        selection = ChannelSelection() if self.force_mount_all else self.mount_last
        channels_set = set(selection.select_from(app.gateway.channels_model))

        # Bind every wire
        for name in channels_set:
            # Install on router
            app.add_last_api(name)

        app.add_last_available_channels(channels_set)

    def _mount_next(self, app: GatewayWebApp) -> None:
        selection = ChannelSelection() if self.force_mount_all else self.mount_next
        channels_set = set(selection.select_from(app.gateway.channels_model))

        # Bind every wire
        for name in channels_set:
            # Install on router
            app.add_next_api(name)

        app.add_next_available_channels(channels_set)

    def _mount_send(self, app: GatewayWebApp) -> None:
        selection = ChannelSelection() if self.force_mount_all else self.mount_send
        channels_set = set(selection.select_from(app.gateway.channels_model))
        seen_channels = set()

        # Bind every wire
        if self.force_mount_all:
            for channel in channels_set:
                app.add_send_api(channel)
            seen_channels = channels_set

        else:
            for channel, _ in app.gateway.channels._send_channels.keys():
                if channel in seen_channels or channel not in channels_set:
                    continue
                seen_channels.add(channel)
                # Install on router
                app.add_send_api(channel)

            missing_channels = channels_set - seen_channels
            if missing_channels:
                log.info(f"Requested channels missing send routes are: {list(missing_channels)}")

        app.add_send_available_channels(seen_channels)

    def _mount_state(self, app: GatewayWebApp) -> None:
        selection = ChannelSelection() if self.force_mount_all else self.mount_state
        channels_set = set(selection.select_from(app.gateway.channels_model, state_channels=True))
        seen_channels = set()

        # Bind every wire
        if self.force_mount_all:
            for state_channel in channels_set:
                app.add_state_api(state_channel)
            seen_channels = channels_set

        else:
            for state_channel, _ in app.gateway.channels._state_requests.keys():
                if state_channel in seen_channels or state_channel not in channels_set:
                    continue
                seen_channels.add(state_channel)
                # Install on router
                app.add_state_api(state_channel)

            missing_channels = channels_set - seen_channels
            if missing_channels:
                log.info(f"Requested channels missing state routes: {list(channel[2:] for channel in missing_channels)}")

        app.add_state_available_channels(seen_channels)

    def _mount_lookup(self, app: GatewayWebApp) -> None:
        selection = ChannelSelection() if self.force_mount_all else self.mount_lookup
        channels_set = set(selection.select_from(app.gateway.channels_model))

        # Bind every wire
        for name in selection.select_from(app.gateway.channels_model):
            # Install on router
            app.add_lookup_api(name)

        app.add_lookup_available_channels(channels_set)



from datetime import timedelta
from typing import Dict, List, Optional, Type

import csp
import numpy as np
from csp import Enum, ts
from csp.typing import Numpy1DArray

from csp_gateway import (
    Channels,
    ChannelSelection,
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewayStruct,
    State,
)

__all__ = (
    "MyEnum",
    "MyStruct",
    "MyGatewayChannels",
    "MySmallGatewayChannels",
    "MySmallGateway",
    "MyGateway",
    "MySetModule",
    "MyGetModule",
    "MyExampleModule",
    "MyDictBasketModule",
    "MyNoTickDictBasket",
)


class MyEnum(Enum):
    ONE = 1
    TWO = 2


class MyStruct(GatewayStruct):
    foo: float
    time: timedelta
    my_flag: bool = True
    _new: bool = True


class MyGatewayChannels(GatewayChannels):
    my_static: float = 0.0
    my_static_dict: Dict[str, float] = {}
    my_static_list: List[str] = []
    my_channel: ts[MyStruct] = None
    s_my_channel: ts[State[MyStruct]] = None
    my_list_channel: ts[List[MyStruct]] = None
    s_my_list_channel: ts[State[MyStruct]] = None
    my_enum_basket: Dict[MyEnum, ts[MyStruct]] = None
    my_str_basket: Dict[str, ts[MyStruct]] = None
    my_enum_basket_list: Dict[MyEnum, ts[List[MyStruct]]] = None
    my_array_channel: ts[Numpy1DArray[float]] = None


class MySmallGatewayChannels(GatewayChannels):
    example: ts[int] = None
    s_example: ts[State[int]] = None
    my_str_basket: Dict[str, ts[float]] = None


class MySmallGateway(Gateway):
    channels_model: Type[Channels] = MySmallGatewayChannels


class MyGateway(Gateway):
    channels_model: Type[Channels] = MyGatewayChannels  # type: ignore[assignment]


class MySetModule(GatewayModule):
    my_data: ts[MyStruct]
    my_data2: ts[MyStruct]
    my_list_data: ts[List[MyStruct]]
    by_key: bool = True

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: ["my_key", "my_key2"]}

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        channels.set_channel(MyGatewayChannels.my_channel, self.my_data)
        channels.set_channel(
            MyGatewayChannels.my_list_channel,
            self.my_list_data,
        )
        channels.add_send_channel(MyGatewayChannels.my_channel)
        channels.set_state(MyGatewayChannels.my_channel, "id")
        channels.add_send_channel(MyGatewayChannels.my_list_channel)
        channels.set_state(MyGatewayChannels.my_list_channel, "id")

        channels.set_channel(MyGatewayChannels.my_array_channel, csp.const(np.array([1.0, 2.0])))

        if self.by_key:
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data, MyEnum.ONE)
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data2, MyEnum.TWO)
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data, "my_key")
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data2, "my_key2")
            channels.set_channel(MyGatewayChannels.my_enum_basket_list, self.my_list_data, MyEnum.ONE)
            channels.set_channel(MyGatewayChannels.my_enum_basket_list, self.my_list_data, MyEnum.TWO)
        else:
            channels.set_channel(
                MyGatewayChannels.my_enum_basket,
                {
                    MyEnum.ONE: self.my_data,
                    MyEnum.TWO: self.my_data2,
                },
            )
            channels.set_channel(
                MyGatewayChannels.my_str_basket,
                {
                    "my_key": self.my_data,
                    "my_key2": self.my_data2,
                },
            )
            channels.set_channel(
                MyGatewayChannels.my_enum_basket_list,
                {
                    MyEnum.ONE: self.my_list_data,
                    MyEnum.TWO: self.my_list_data,
                },
            )


class MyGetModule(GatewayModule):
    requires: Optional[ChannelSelection] = []

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        csp.add_graph_output("my_channel", channels.get_channel(MyGatewayChannels.my_channel))
        csp.add_graph_output("my_list_channel", channels.get_channel(MyGatewayChannels.my_list_channel))
        csp.add_graph_output("my_array_channel", channels.get_channel(MyGatewayChannels.my_array_channel))
        # Get by basket
        for k, v in channels.get_channel(MyGatewayChannels.my_enum_basket).items():
            csp.add_graph_output(f"my_enum_basket[{k}]", v)
        for k, v in channels.get_channel(MyGatewayChannels.my_str_basket).items():
            csp.add_graph_output(f"my_str_basket[{k}]", v)
        for k, v in channels.get_channel(MyGatewayChannels.my_enum_basket_list).items():
            csp.add_graph_output(f"my_enum_basket_list[{k}]", v)

        # Get by indexer
        csp.add_graph_output(
            "my_enum_basket_ONE",
            channels.get_channel(MyGatewayChannels.my_enum_basket, MyEnum.ONE),
        )
        csp.add_graph_output(
            "my_enum_basket_TWO",
            channels.get_channel(MyGatewayChannels.my_enum_basket, MyEnum.TWO),
        )
        csp.add_graph_output(
            "my_str_basket_my_key",
            channels.get_channel(MyGatewayChannels.my_str_basket, "my_key"),
        )
        csp.add_graph_output(
            "my_str_basket_my_key2",
            channels.get_channel(MyGatewayChannels.my_str_basket, "my_key2"),
        )


class MyExampleModule(GatewayModule):
    interval: timedelta = timedelta(seconds=1)

    @csp.node
    def subscribe(
        self,
        trigger: ts[bool],
    ) -> ts[int]:
        with csp.state():
            last_x = 0
        if csp.ticked(trigger):
            last_x += 1
            return last_x

    def connect(self, channels: MySmallGatewayChannels):
        channels.set_channel(
            MySmallGatewayChannels.example,
            self.subscribe(csp.timer(interval=self.interval, value=True)),
        )
        channels.add_send_channel(MySmallGatewayChannels.example)
        channels.set_state(MySmallGatewayChannels.example, "id")


class MyDictBasketModule(GatewayModule):
    my_data: ts[float]

    def dynamic_keys(self):
        return {MySmallGatewayChannels.my_str_basket: ["my_key"]}

    def connect(self, channels: MySmallGatewayChannels):
        channels.set_channel(MySmallGatewayChannels.my_str_basket, {"my_key": self.my_data})
        for k, v in channels.get_channel(MySmallGatewayChannels.my_str_basket).items():
            csp.add_graph_output(f"my_str_basket[{k}]", v)


class MyNoTickDictBasket(GatewayModule):
    my_data: ts[float]

    def dynamic_keys(self):
        return {MySmallGatewayChannels.my_str_basket: ["my_key"]}

    @csp.node
    def subscribe(
        self,
        trigger: ts[bool],
        keys: List[str],
    ) -> csp.OutputBasket(Dict[str, ts[float]], shape="keys"):
        if csp.ticked(trigger):
            csp.output({"my_key": self.my_data})

    def connect(self, channels: MySmallGatewayChannels):
        channels.set_channel(
            MySmallGatewayChannels.my_str_basket,
            self.subscribe(
                csp.timer(interval=timedelta(seconds=5), value=True),
                self.dynamic_keys()[MySmallGatewayChannels.my_str_basket],
            ),
        )
        for k, v in channels.get_channel(MySmallGatewayChannels.my_str_basket).items():
            csp.add_graph_output(f"my_str_basket[{k}]", v)



from typing import Optional

import csp
from csp.impl.types.tstype import isTsType
from pydantic import Field

from csp_gateway.server import ChannelSelection, ChannelsType, GatewayModule
from csp_gateway.utils import is_dict_basket

__all__ = ("AddChannelsToGraphOutput",)


class AddChannelsToGraphOutput(GatewayModule):
    selection: ChannelSelection = Field(default_factory=ChannelSelection)
    requires: Optional[ChannelSelection] = []

    def connect(self, channels: ChannelsType):
        for field in self.selection.select_from(channels):
            outer_type = channels.get_outer_type(field)
            # list baskets not supported yet
            if is_dict_basket(outer_type):
                edge = channels.get_channel(field)
                for k, v in edge.items():
                    csp.add_graph_output(f"{field}[{k}]", v)
            elif isTsType(outer_type):
                edge = channels.get_channel(field)
                csp.add_graph_output(f"{field}", edge)



from enum import Enum


class ReadWriteMode(str, Enum):
    """Enum representing whether a component is set to read, write, or both."""

    READ = "READ"
    WRITE = "WRITE"
    READ_AND_WRITE = "READ_AND_WRITE"



import logging

import numpy
import pandas
import polars
import pytest
from packaging import version

from csp_gateway import ClientConfig, GatewayClient, ResponseWrapper
from csp_gateway.client.client import _host
from csp_gateway.utils import get_thread

#  Struct for response wrapper
#  class MyTypeStruct(BaseModel):
#      d_str: str
#      d_int: int
#      d_float: float
#      d_datetime: datetime
#      d_int_or_float: Union[int, float]

list_data = [
    ("integer", [1, 2, 3, 4, 5]),
    ("string", ["data1", "data2", "data3", "data4", "data5"]),
]

data = [
    {
        "d_str": "A",
        "d_int": 1,
        "d_float": 1.0,
        "d_datetime": "2024-01-01 00:00:00+00:00",
        "d_int_or_float": 1,
    },
    {
        "d_str": "A",
        "d_int": 1,
        "d_float": 1.0,
        "d_datetime": "2024-01-01 00:00:00+00:00",
        "d_int_or_float": 1.0,
    },
    {
        "d_int": 1,
        "d_float": 1.0,
        "d_datetime": "2024-01-01 00:00:00+00:00",
        "d_int_or_float": 1,
    },
    # NOTE: as_pandas_df cannot handle unset int values
    #  {'d_str': 'A', 'd_float': 1.0, 'd_datetime': "2024-01-01 00:00:00+00:00", 'd_int_or_float': 1},
    {
        "d_str": "A",
        "d_int": 1,
        "d_datetime": "2024-01-01 00:00:00+00:00",
        "d_int_or_float": 1,
    },
    {"d_str": "A", "d_int": 1, "d_float": 1.0, "d_int_or_float": 1},
    {
        "d_str": "A",
        "d_int": 1,
        "d_float": 1.0,
        "d_datetime": "2024-01-01 00:00:00+00:00",
    },
]
schema = {
    "properties": {
        "d_str": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "d_str"},
        "d_datetime": {
            "anyOf": [{"type": "string", "format": "date-time"}, {"type": "null"}],
            "title": "d_datetime",
        },
        "d_int": {
            "anyOf": [{"type": "integer"}, {"type": "null"}],
            "title": "d_intArea",
        },
        "d_float": {
            "anyOf": [{"type": "number"}, {"type": "null"}],
            "title": "d_float",
        },
        "d_enum": {
            "anyOf": [
                {
                    "type": "string",
                    "enum": ["UNKNOWN", "ENUM_1", "ENUM_2"],
                    "title": "d_enum",
                },
                {"type": "null"},
            ],
            "title": "d_enum",
            "default": "UNKNOWN",
        },
        "d_int_or_float": {
            "anyOf": [{"type": "number"}, {"type": "integer"}, {"type": "null"}],
            "title": "Quantity",
            "default": 0,
        },
    },
    "additionalProperties": False,
    "type": "object",
    "title": "MyTypeStruct",
}


@pytest.mark.skipif(
    version.parse(numpy.__version__) < version.parse("1.25"),
    reason="NumPy version less than 1.25 does not have numpy.dtypes",
)
def test_response_wrapper_empty_pandas():
    resp = ResponseWrapper(json_data={}, openapi_schema=schema)
    # Test as_pandas_df
    assert resp.as_json() == {}
    assert isinstance(resp.as_pandas_df().dtypes["d_str"], pandas.core.arrays.string_.StringDtype)
    assert isinstance(resp.as_pandas_df().dtypes["d_int"], numpy.dtypes.Int64DType)
    assert isinstance(resp.as_pandas_df().dtypes["d_float"], numpy.dtypes.Float64DType)
    assert isinstance(resp.as_pandas_df().dtypes["d_datetime"], pandas.core.arrays.string_.StringDtype)
    assert isinstance(resp.as_pandas_df().dtypes["d_int_or_float"], numpy.dtypes.ObjectDType)


def test_response_wrapper_empty_polars():
    resp = ResponseWrapper(json_data={}, openapi_schema=schema)
    # Test as_polars_df
    assert resp.as_polars_df().schema["d_str"] == polars.Utf8
    assert resp.as_polars_df().schema["d_int"] == polars.Int64
    assert resp.as_polars_df().schema["d_float"] == polars.Float64
    assert resp.as_polars_df().schema["d_datetime"] == polars.Utf8
    assert resp.as_polars_df().schema["d_int_or_float"] == polars.Object


def test_get_event_loop_off_thread(caplog):
    caplog.set_level(logging.DEBUG)

    bad_log = "THIS LOG MESSAGE PROBABLY WOULD NEVER NATURALLY OCCUR"

    def instantiate_in_thread():
        try:
            GatewayClient()
        except Exception:
            # This should not be hit
            logging.error(bad_log)

    thread = get_thread(target=instantiate_in_thread)
    thread.start()
    thread.join()

    expected_message = "Attempted to get event loop without one running, creating a new one..."
    assert any(expected_message in record.message for record in caplog.records)
    assert all(bad_log not in record.message for record in caplog.records)


def test_response_wrapper():
    resp = ResponseWrapper(json_data=data, openapi_schema=schema)
    assert resp.as_json() == data

    # Test Dict Data
    pandas_df = resp.as_pandas_df()
    for idx, df_row in pandas_df.iterrows():
        for col in pandas_df.columns:
            if pandas.isnull(df_row[col]):
                assert col not in data[idx]
            else:
                assert df_row[col] == data[idx][col]

    polars_df = resp.as_polars_df()
    for idx in range(len(polars_df)):
        for col in polars_df.columns:
            if polars_df[col][idx] is None:
                assert col not in data[idx]
            else:
                assert polars_df[col][idx] == data[idx][col]

    # Test List Data
    for typ, lst in list_data:
        lst_schema = {"type": typ}
        lst_resp = ResponseWrapper(json_data=lst, openapi_schema=lst_schema)
        assert lst_resp.as_json() == lst

        pandas_df = lst_resp.as_pandas_df()
        assert pandas_df.iloc[:, 0].tolist() == lst

        polars_df = lst_resp.as_polars_df()
        assert polars_df[:, 0].to_list() == lst


def test_client_instantiation_convenenience():
    client = GatewayClient()
    assert client.config.host == "localhost"
    assert client.config.port == 8000
    assert client.config.protocol == "http"

    client = GatewayClient(host="https://my.test.gateway", port=None)
    assert client.config.host == "my.test.gateway"
    assert client.config.port is None
    assert client.config.protocol == "https"


def test_host_parsing():
    cfg = ClientConfig(host="localhost", port=8000, protocol="http")
    assert _host(cfg) == "http://localhost:8000"

    cfg = ClientConfig(host="https://my.host", port=None, protocol="http")
    assert _host(cfg) == "https://my.host"



from typing import Any, Callable, Coroutine, Dict, Union

from fastapi.exceptions import RequestErrorModel

__all__ = (
    "Error404",
    "get_default_responses",
)

NoArgsNoReturnFuncT = Callable[[], None]
NoArgsNoReturnAsyncFuncT = Callable[[], Coroutine[Any, Any, None]]
NoArgsNoReturnDecorator = Callable[[Union[NoArgsNoReturnFuncT, NoArgsNoReturnAsyncFuncT]], NoArgsNoReturnAsyncFuncT]


class Error404(RequestErrorModel):  # type: ignore[misc, valid-type]
    detail: str = ""


def get_default_responses() -> Dict[Union[int, str], Dict[str, Any]]:
    return {404: {"model": Error404}}
